# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import collections.abc
import pickle
from io import StringIO

import pyomo.common.unittest as unittest
import pyomo.kernel as pmo
from pyomo.common.log import LoggingIntercept
from pyomo.core.kernel.base import ICategorizedObject, ICategorizedObjectContainer
from pyomo.core.kernel.homogeneous_container import IHomogeneousContainer
from pyomo.core.kernel.list_container import ListContainer
from pyomo.core.kernel.block import block, block_list

#
# There are no fully implemented test suites in this
# file. These classes are meant to be used to test
# full implementations of ListContainer containers.
# To test a ListContainer implementation, just import
# this class and use it as a subclass in another test suite.
#

# Note: we need to test with a pickle protocol
#       that is high enough to support __slots__
#       and weakref (bas
_pickle_test_protocol = pickle.HIGHEST_PROTOCOL


class _bad_ctype:
    ctype = "_this_is_definitely_not_the_ctype_being_tested"


class _TestListContainerBase:
    # set by derived class
    _container_type = None
    _ctype_factory = None

    def test_overwrite_warning(self):
        c = self._container_type()
        out = StringIO()
        with LoggingIntercept(out, 'pyomo.core'):
            c.append(self._ctype_factory())
            c[0] = c[0]
        assert out.getvalue() == ""
        with LoggingIntercept(out, 'pyomo.core'):
            c[0] = self._ctype_factory()
        assert out.getvalue() == (
            "Implicitly replacing the entry [0] "
            "(type=%s) with a new object (type=%s). "
            "This is usually indicative of a modeling "
            "error. To avoid this warning, delete the "
            "original object from the container before "
            "assigning a new object.\n"
            % (
                self._ctype_factory().__class__.__name__,
                self._ctype_factory().__class__.__name__,
            )
        )

    def test_ctype(self):
        c = self._container_type()
        ctype = self._ctype_factory().ctype
        self.assertIs(c.ctype, ctype)
        self.assertIs(type(c)._ctype, ctype)
        self.assertIs(self._container_type._ctype, ctype)

    def test_init1(self):
        clist = self._container_type()

    def test_init2(self):
        index = range(5)
        c = self._container_type(self._ctype_factory() for i in index)
        with self.assertRaises(TypeError):
            d = self._container_type(*tuple(self._ctype_factory() for i in index))

    def test_type(self):
        clist = self._container_type()
        self.assertTrue(isinstance(clist, ICategorizedObject))
        self.assertTrue(isinstance(clist, ICategorizedObjectContainer))
        self.assertTrue(isinstance(clist, IHomogeneousContainer))
        self.assertTrue(isinstance(clist, ListContainer))
        self.assertTrue(isinstance(clist, collections.abc.Sequence))
        self.assertTrue(issubclass(type(clist), collections.abc.Sequence))
        self.assertTrue(isinstance(clist, collections.abc.MutableSequence))
        self.assertTrue(issubclass(type(clist), collections.abc.MutableSequence))

    def test_len1(self):
        c = self._container_type()
        self.assertEqual(len(c), 0)

    def test_len2(self):
        index = range(5)
        c = self._container_type(self._ctype_factory() for i in index)
        self.assertEqual(len(c), len(index))

    def test_append(self):
        c = self._container_type()
        index = range(5)
        self.assertEqual(len(c), 0)
        for i in index:
            c_new = self._ctype_factory()
            c.append(c_new)
            self.assertEqual(id(c[-1]), id(c_new))
            self.assertEqual(len(c), i + 1)

    def test_insert(self):
        c = self._container_type()
        index = range(5)
        self.assertEqual(len(c), 0)
        for i in index:
            c_new = self._ctype_factory()
            c.insert(0, c_new)
            self.assertEqual(id(c[0]), id(c_new))
            self.assertEqual(len(c), i + 1)

    def test_setitem(self):
        c = self._container_type()
        index = range(5)
        for i in index:
            c.append(self._ctype_factory())
        for i in index:
            c_new = self._ctype_factory()
            self.assertNotEqual(id(c_new), id(c[i]))
            c[i] = c_new
            self.assertEqual(len(c), len(index))
            self.assertEqual(id(c_new), id(c[i]))

    def test_wrong_type_init(self):
        index = range(5)
        with self.assertRaises(TypeError):
            c = self._container_type(_bad_ctype() for i in index)

    def test_wrong_type_append(self):
        c = self._container_type()
        c.append(self._ctype_factory())
        with self.assertRaises(TypeError):
            c.append(_bad_ctype())

    def test_wrong_type_insert(self):
        c = self._container_type()
        c.append(self._ctype_factory())
        c.insert(0, self._ctype_factory())
        with self.assertRaises(TypeError):
            c.insert(0, _bad_ctype())

    def test_wrong_type_setitem(self):
        c = self._container_type()
        c.append(self._ctype_factory())
        c[0] = self._ctype_factory()
        with self.assertRaises(TypeError):
            c[0] = _bad_ctype()

    def test_has_parent_init(self):
        c = self._container_type()
        c.append(self._ctype_factory())
        with self.assertRaises(ValueError):
            c.append(c[0])
        with self.assertRaises(ValueError):
            d = self._container_type(c)

    def test_has_parent_append(self):
        c = self._container_type()
        c.append(self._ctype_factory())
        with self.assertRaises(ValueError):
            c.append(c[0])
        d = []
        d.append(c[0])
        d = self._container_type()
        with self.assertRaises(ValueError):
            d.append(c[0])

    def test_has_parent_insert(self):
        c = self._container_type()
        c.append(self._ctype_factory())
        c.insert(0, self._ctype_factory())
        with self.assertRaises(ValueError):
            c.insert(0, c[0])
        d = []
        d.insert(0, c[0])
        d = self._container_type()
        with self.assertRaises(ValueError):
            d.insert(0, c[0])

    def test_has_parent_setitem(self):
        c = self._container_type()
        c.append(self._ctype_factory())
        c[0] = self._ctype_factory()
        c[0] = c[0]
        c.append(self._ctype_factory())
        with self.assertRaises(ValueError):
            c[0] = c[1]

    def test_setitem_exists_overwrite(self):
        index = range(5)
        c = self._container_type(self._ctype_factory() for i in index)
        self.assertEqual(len(c), len(index))
        for i in index:
            cdata = c[i]
            self.assertEqual(id(cdata.parent), id(c))
            c[i] = self._ctype_factory()
            self.assertEqual(len(c), len(index))
            self.assertNotEqual(id(cdata), id(c[i]))
            self.assertEqual(cdata.parent, None)

    def test_delitem(self):
        index = range(5)
        c = self._container_type(self._ctype_factory() for i in index)
        self.assertEqual(len(c), len(index))
        for i in index:
            cdata = c[0]
            self.assertEqual(id(cdata.parent), id(c))
            del c[0]
            self.assertEqual(len(c), len(index) - (i + 1))
            self.assertEqual(cdata.parent, None)

    def test_iter(self):
        index = range(5)
        c = self._container_type(self._ctype_factory() for i in index)
        self.assertEqual(len(c), len(index))
        raw_list = c[:]
        self.assertEqual(type(raw_list), list)
        for c1, c2 in zip(raw_list, c):
            self.assertEqual(id(c1), id(c2))

    def test_reverse(self):
        index = range(5)
        c = self._container_type(self._ctype_factory() for i in index)
        raw_list = c[:]
        self.assertEqual(type(raw_list), list)
        for c1, c2 in zip(reversed(c), reversed(raw_list)):
            self.assertEqual(id(c1), id(c2))
        c.reverse()
        raw_list.reverse()
        for c1, c2 in zip(c, raw_list):
            self.assertEqual(id(c1), id(c2))

    def test_remove(self):
        model = block()
        index = range(5)
        c = self._container_type(self._ctype_factory() for i in index)
        for i in index:
            cdata = c[0]
            self.assertEqual(cdata in c, True)
            c.remove(cdata)
            self.assertEqual(cdata in c, False)

    def test_pop(self):
        index = range(5)
        c = self._container_type(self._ctype_factory() for i in index)
        for i in index:
            cdata = c[-1]
            self.assertEqual(cdata in c, True)
            last = c.pop()
            self.assertEqual(cdata in c, False)
            self.assertEqual(id(cdata), id(last))

    def test_index(self):
        index = range(5)
        c = self._container_type(self._ctype_factory() for i in index)
        for i in index:
            cdata = c[i]
            self.assertEqual(c.index(cdata), i)
            self.assertEqual(c.index(cdata, start=i), i)
            with self.assertRaises(ValueError):
                c.index(cdata, start=i + 1)
            with self.assertRaises(ValueError):
                c.index(cdata, start=i, stop=i)
            with self.assertRaises(ValueError):
                c.index(cdata, stop=i)
            self.assertEqual(c.index(cdata, start=i, stop=i + 1), i)
            with self.assertRaises(ValueError):
                c.index(cdata, start=i + 1, stop=i + 1)
            self.assertEqual(c.index(cdata, start=-len(index) + i), i)
            if i == index[-1]:
                self.assertEqual(c.index(cdata, start=-len(index) + i + 1), i)
            else:
                with self.assertRaises(ValueError):
                    self.assertEqual(c.index(cdata, start=-len(index) + i + 1), i)
            if i == index[-1]:
                with self.assertRaises(ValueError):
                    self.assertEqual(c.index(cdata, stop=-len(index) + i + 1), i)
            else:
                self.assertEqual(c.index(cdata, stop=-len(index) + i + 1), i)
        tmp = self._ctype_factory()
        with self.assertRaises(ValueError):
            c.index(tmp)
        with self.assertRaises(ValueError):
            c.index(tmp, stop=len(c) + 1)

    def test_extend(self):
        index = range(5)
        c = self._container_type(self._ctype_factory() for i in index)
        c_more_list = [self._ctype_factory() for i in index]
        self.assertEqual(len(c), len(index))
        self.assertTrue(len(c_more_list) > 0)
        for cdata in c_more_list:
            self.assertEqual(cdata.parent, None)
        c.extend(c_more_list)
        for cdata in c_more_list:
            self.assertEqual(id(cdata.parent), id(c))

    def test_count(self):
        index = range(5)
        c = self._container_type(self._ctype_factory() for i in index)
        for i in index:
            self.assertEqual(c.count(c[i]), 1)

    def test_pickle(self):
        index = range(5)
        clist = self._container_type(self._ctype_factory() for i in index)
        clist.append(self._container_type())
        index = list(index)
        index = index + [len(index)]
        for i in index:
            self.assertTrue(clist[i].parent is clist)
        pickled_clist = pickle.loads(
            pickle.dumps(clist, protocol=_pickle_test_protocol)
        )
        self.assertTrue(isinstance(pickled_clist, self._container_type))
        self.assertTrue(pickled_clist.parent is None)
        self.assertEqual(len(pickled_clist), len(index))
        self.assertNotEqual(id(pickled_clist), id(clist))
        for i in index:
            self.assertNotEqual(id(pickled_clist[i]), id(clist[i]))
            self.assertTrue(pickled_clist[i].parent is pickled_clist)
            self.assertTrue(clist[i].parent is clist)

    def test_eq(self):
        clist1 = self._container_type()
        clist1.append(self._ctype_factory())
        clist2 = self._container_type()
        clist2.append(self._ctype_factory())

        self.assertNotEqual(clist1, set())
        self.assertFalse(clist1 == set())
        self.assertNotEqual(clist1, list())
        self.assertFalse(clist1 == list())
        self.assertNotEqual(clist1, tuple())
        self.assertFalse(clist1 == tuple())
        self.assertNotEqual(clist1, dict())
        self.assertFalse(clist1 == dict())

        self.assertTrue(clist1 == clist1)
        self.assertEqual(clist1, clist1)
        self.assertTrue(clist1 == list(clist1))
        self.assertEqual(clist1, list(clist1))
        self.assertTrue(clist1 == tuple(clist1))
        self.assertEqual(clist1, tuple(clist1))
        self.assertTrue(list(clist1) == clist1)
        self.assertEqual(list(clist1), clist1)
        self.assertTrue(tuple(clist1) == clist1)
        self.assertEqual(tuple(clist1), clist1)

        self.assertFalse(clist2 == clist1)
        self.assertTrue(clist2 != clist1)
        self.assertNotEqual(clist2, clist1)

        self.assertFalse(clist1 == clist2)
        self.assertTrue(clist1 != clist2)
        self.assertNotEqual(clist1, clist2)

        clist1 = self._container_type()

        self.assertEqual(clist1, set())
        self.assertTrue(clist1 == set())
        self.assertEqual(clist1, list())
        self.assertTrue(clist1 == list())
        self.assertEqual(clist1, tuple())
        self.assertTrue(clist1 == tuple())
        self.assertNotEqual(clist1, dict())
        self.assertFalse(clist1 == dict())

    def test_child(self):
        clist = self._container_type()
        c = self._ctype_factory()
        clist.append(c)
        with self.assertRaises(KeyError):
            clist.child(1)
        self.assertIs(clist.child(0), c)

    def test_name(self):
        children = [self._ctype_factory() for i in range(5)]
        children.append(self._container_type())
        children[-1].append(self._ctype_factory())

        for c in children:
            self.assertTrue(c.parent is None)
            self.assertEqual(c.local_name, None)
            self.assertEqual(c.name, None)

        clist = self._container_type()
        self.assertTrue(clist.parent is None)
        self.assertEqual(clist.local_name, None)
        self.assertEqual(clist.name, None)
        clist.extend(children)
        names = pmo.generate_names(clist)
        for i, c in enumerate(children):
            self.assertTrue(c.parent is clist)
            self.assertEqual(c.local_name, "[%s]" % (i))
            self.assertEqual(c.name, "[%s]" % (i))
            self.assertEqual(c.name, names[c])
        for c in clist.components():
            self.assertNotEqual(c.name, None)
            self.assertEqual(c.name, names[c])

        model = block()
        model.clist = clist
        self.assertTrue(model.parent is None)
        self.assertTrue(clist.parent is model)
        self.assertEqual(clist.local_name, "clist")
        self.assertEqual(clist.name, "clist")
        names = pmo.generate_names(model)
        for i, c in enumerate(children):
            self.assertTrue(c.parent is clist)
            self.assertEqual(c.local_name, "[%s]" % (i))
            self.assertEqual(c.name, "clist[%s]" % (i))
            self.assertEqual(c.name, names[c])
        for c in clist.components():
            self.assertNotEqual(c.name, None)
            self.assertEqual(c.name, names[c])

        b = block()
        b.model = model
        self.assertTrue(b.parent is None)
        self.assertTrue(model.parent is b)
        self.assertTrue(clist.parent is model)
        self.assertEqual(clist.local_name, "clist")
        self.assertEqual(clist.name, "model.clist")
        names = pmo.generate_names(b)
        for i, c in enumerate(children):
            self.assertTrue(c.parent is clist)
            self.assertEqual(c.local_name, "[%s]" % (i))
            self.assertEqual(c.name, "model.clist[%s]" % (i))
            self.assertEqual(c.name, names[c])
        for c in clist.components():
            self.assertNotEqual(c.name, None)
            self.assertEqual(c.name, names[c])

        blist = block_list()
        blist.append(b)
        self.assertTrue(blist.parent is None)
        self.assertTrue(b.parent is blist)
        self.assertTrue(model.parent is b)
        self.assertTrue(clist.parent is model)
        self.assertEqual(clist.local_name, "clist")
        self.assertEqual(clist.name, "[0].model.clist")
        for i, c in enumerate(children):
            self.assertTrue(c.parent is clist)
            self.assertEqual(c.local_name, "[%s]" % (i))
            self.assertEqual(c.name, "[0].model.clist[%s]" % (i))

        m = block()
        m.blist = blist
        self.assertTrue(m.parent is None)
        self.assertTrue(blist.parent is m)
        self.assertTrue(b.parent is blist)
        self.assertTrue(model.parent is b)
        self.assertTrue(clist.parent is model)
        self.assertEqual(clist.local_name, "clist")
        self.assertEqual(clist.name, "blist[0].model.clist")
        names = pmo.generate_names(m)
        for i, c in enumerate(children):
            self.assertTrue(c.parent is clist)
            self.assertEqual(c.local_name, "[%s]" % (i))
            self.assertEqual(c.name, "blist[0].model.clist[%s]" % (i))
            self.assertEqual(c.name, names[c])
        for c in clist.components():
            self.assertNotEqual(c.name, None)
            self.assertEqual(c.name, names[c])
        names = pmo.generate_names(m)
        for c in m.children():
            self.assertEqual(c.name, names[c])

    def test_components(self):
        clist = self._container_type()
        self.assertEqual(list(clist.components()), [])

        clistflattened = []
        clistflattened.append(self._ctype_factory())
        clist.append(clistflattened[-1])
        clistflattened.append(self._ctype_factory())
        clist.append(clistflattened[-1])
        self.assertEqual(
            list(id(_c) for _c in clist.components()),
            list(id(_c) for _c in clistflattened),
        )

        csublist = self._container_type()
        self.assertEqual(list(csublist.components()), [])

        csublistflattened = []
        csublistflattened.append(self._ctype_factory())
        clistflattened.append(csublistflattened[-1])
        csublist.append(csublistflattened[-1])

        csublistflattened.append(self._ctype_factory())
        clistflattened.append(csublistflattened[-1])
        csublist.append(csublistflattened[-1])

        csublistflattened.append(self._ctype_factory())
        clistflattened.append(csublistflattened[-1])
        csublist.append(csublistflattened[-1])

        self.assertEqual(
            list(id(_c) for _c in csublist.components()),
            list(id(_c) for _c in csublistflattened),
        )
        self.assertEqual(
            len(set(id(_c) for _c in csublist.components())),
            len(list(id(_c) for _c in csublist.components())),
        )
        self.assertEqual(len(set(id(_c) for _c in csublist.components())), 3)

        clist.append(csublist)
        self.assertEqual(
            list(id(_c) for _c in clist.components()),
            list(id(_c) for _c in clistflattened),
        )
        self.assertEqual(
            len(set(id(_c) for _c in clist.components())),
            len(list(id(_c) for _c in clist.components())),
        )
        self.assertEqual(len(set(id(_c) for _c in clist.components())), 5)

    def test_preorder_traversal(self):
        traversal = []
        clist = self._container_type()
        traversal.append(clist)
        clist.append(self._ctype_factory())
        traversal.append(clist[-1])
        clist.append(self._container_type())
        traversal.append(clist[-1])
        clist[1].append(self._ctype_factory())
        traversal.append(clist[1][-1])
        clist.append(self._ctype_factory())
        traversal.append(clist[-1])

        descend = lambda x: not x._is_heterogeneous_container

        self.assertEqual(
            [c.name for c in traversal],
            [c.name for c in pmo.preorder_traversal(clist, descend=descend)],
        )
        self.assertEqual(
            [id(c) for c in traversal],
            [id(c) for c in pmo.preorder_traversal(clist, descend=descend)],
        )

        return clist, traversal

    def test_preorder_traversal_descend_check(self):
        traversal = []
        clist = self._container_type()
        traversal.append(clist)
        clist.append(self._ctype_factory())
        traversal.append(clist[-1])
        clist.append(self._container_type())
        traversal.append(clist[-1])
        clist[1].append(self._ctype_factory())
        traversal.append(clist[1][-1])
        clist.append(self._ctype_factory())
        traversal.append(clist[-1])

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return False

        descend.seen = []
        order = list(pmo.preorder_traversal(clist, descend=descend))
        self.assertEqual(len(order), 1)
        self.assertIs(order[0], clist)
        self.assertEqual(len(descend.seen), 1)
        self.assertIs(descend.seen[0], clist)

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return not x._is_heterogeneous_container

        descend.seen = []
        order = list(pmo.preorder_traversal(clist, descend=descend))
        self.assertEqual([c.name for c in traversal], [c.name for c in order])
        self.assertEqual([id(c) for c in traversal], [id(c) for c in order])
        self.assertEqual(
            [c.name for c in traversal if c._is_container],
            [c.name for c in descend.seen],
        )
        self.assertEqual(
            [id(c) for c in traversal if c._is_container], [id(c) for c in descend.seen]
        )

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return not x._is_heterogeneous_container

        descend.seen = []
        order = list(pmo.preorder_traversal(clist, descend=descend))
        self.assertEqual([c.name for c in traversal], [c.name for c in order])
        self.assertEqual([id(c) for c in traversal], [id(c) for c in order])
        self.assertEqual(
            [c.name for c in traversal if c._is_container],
            [c.name for c in descend.seen],
        )
        self.assertEqual(
            [id(c) for c in traversal if c._is_container], [id(c) for c in descend.seen]
        )
        return clist, traversal


class _TestActiveListContainerBase(_TestListContainerBase):
    def test_active_type(self):
        clist = self._container_type()
        self.assertTrue(isinstance(clist, ICategorizedObject))
        self.assertTrue(isinstance(clist, ICategorizedObjectContainer))
        self.assertTrue(isinstance(clist, IHomogeneousContainer))
        self.assertTrue(isinstance(clist, ListContainer))
        self.assertTrue(isinstance(clist, collections.abc.Sequence))
        self.assertTrue(issubclass(type(clist), collections.abc.Sequence))
        self.assertTrue(isinstance(clist, collections.abc.MutableSequence))
        self.assertTrue(issubclass(type(clist), collections.abc.MutableSequence))

    def test_active(self):
        index = list(range(4))
        clist = self._container_type(self._ctype_factory() for i in index)
        with self.assertRaises(AttributeError):
            clist.active = False
        for c in clist:
            with self.assertRaises(AttributeError):
                c.active = False

        model = block()
        model.clist = clist
        b = block()
        b.model = model
        blist = block_list()
        blist.append(b)
        blist.append(block())
        m = block()
        m.blist = blist

        self.assertEqual(m.active, True)
        self.assertEqual(blist.active, True)
        self.assertEqual(blist[1].active, True)
        self.assertEqual(b.active, True)
        self.assertEqual(model.active, True)
        self.assertEqual(clist.active, True)
        for c in clist:
            self.assertEqual(c.active, True)
        for c in clist.components():
            self.assertEqual(c.active, True)
        for c in clist.components(active=True):
            self.assertEqual(c.active, True)
        self.assertEqual(len(list(clist.components())), len(clist))
        self.assertEqual(
            len(list(clist.components())), len(list(clist.components(active=True)))
        )

        m.deactivate(shallow=False)

        self.assertEqual(m.active, False)
        self.assertEqual(blist.active, False)
        self.assertEqual(blist[1].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(clist.active, False)
        for c in clist:
            self.assertEqual(c.active, False)
        self.assertNotEqual(
            len(list(clist.components())), len(list(clist.components(active=None)))
        )
        self.assertEqual(len(list(clist.components(active=True))), 0)

        test_c = clist[0]
        clist.remove(test_c)
        clist.append(test_c)

        self.assertEqual(m.active, False)
        self.assertEqual(blist.active, False)
        self.assertEqual(blist[1].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(clist.active, False)
        for c in clist:
            self.assertEqual(c.active, False)
        self.assertNotEqual(
            len(list(clist.components())), len(list(clist.components(active=None)))
        )
        self.assertEqual(len(list(clist.components(active=True))), 0)

        clist.remove(test_c)
        test_c.activate()
        self.assertEqual(test_c.active, True)
        self.assertEqual(clist.active, False)
        clist.append(test_c)

        self.assertEqual(m.active, False)
        self.assertEqual(blist.active, False)
        self.assertEqual(blist[1].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(clist.active, False)
        clist.activate()
        self.assertEqual(m.active, False)
        self.assertEqual(blist.active, False)
        self.assertEqual(blist[1].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(clist.active, True)
        for c in clist:
            if c is test_c:
                self.assertEqual(c.active, True)
            else:
                self.assertEqual(c.active, False)
        for c in clist.components():
            if c is test_c:
                self.assertEqual(c.active, True)
            else:
                self.assertEqual(c.active, False)
        for c in clist.components():
            self.assertEqual(c.active, True)
        for c in clist.components(active=True):
            self.assertEqual(c.active, True)
        self.assertNotEqual(
            len(list(clist.components())), len(list(clist.components(active=None)))
        )
        self.assertEqual(len(list(clist.components(active=True))), 1)

        m.activate(shallow=False)

        self.assertEqual(m.active, True)
        self.assertEqual(blist.active, True)
        self.assertEqual(blist[1].active, True)
        self.assertEqual(b.active, True)
        self.assertEqual(model.active, True)
        self.assertEqual(clist.active, True)
        for c in clist:
            self.assertEqual(c.active, True)
        for c in clist.components():
            self.assertEqual(c.active, True)
        for c in clist.components(active=True):
            self.assertEqual(c.active, True)
        self.assertEqual(len(list(clist.components())), len(clist))
        self.assertEqual(
            len(list(clist.components())), len(list(clist.components(active=True)))
        )

        m.deactivate(shallow=False)

        self.assertEqual(m.active, False)
        self.assertEqual(blist.active, False)
        self.assertEqual(blist[1].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(clist.active, False)
        for c in clist:
            self.assertEqual(c.active, False)
        self.assertNotEqual(
            len(list(clist.components())), len(list(clist.components(active=None)))
        )
        self.assertEqual(len(list(clist.components(active=True))), 0)

        clist[len(clist) - 1] = self._ctype_factory()

        self.assertEqual(m.active, False)
        self.assertEqual(blist.active, False)
        self.assertEqual(blist[1].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(clist.active, False)
        clist.activate()
        self.assertEqual(m.active, False)
        self.assertEqual(blist.active, False)
        self.assertEqual(blist[1].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(clist.active, True)
        for i, c in enumerate(clist):
            if i == len(clist) - 1:
                self.assertEqual(c.active, True)
            else:
                self.assertEqual(c.active, False)
        for i, c in enumerate(clist.components(active=None)):
            if i == len(clist) - 1:
                self.assertEqual(c.active, True)
            else:
                self.assertEqual(c.active, False)
        for c in clist.components():
            self.assertEqual(c.active, True)
        for c in clist.components(active=True):
            self.assertEqual(c.active, True)
        self.assertNotEqual(
            len(list(clist.components())), len(list(clist.components(active=None)))
        )
        self.assertEqual(len(list(clist.components(active=True))), 1)

        clist.activate(shallow=False)

        self.assertEqual(m.active, False)
        self.assertEqual(blist.active, False)
        self.assertEqual(blist[1].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(clist.active, True)
        for i, c in enumerate(clist):
            self.assertEqual(c.active, True)
        for c in clist.components():
            self.assertEqual(c.active, True)
        for c in clist.components(active=True):
            self.assertEqual(c.active, True)
        self.assertEqual(len(list(clist.components())), len(clist))
        self.assertEqual(
            len(list(clist.components())), len(list(clist.components(active=True)))
        )

        clist.deactivate(shallow=False)

        self.assertEqual(m.active, False)
        self.assertEqual(blist.active, False)
        self.assertEqual(blist[1].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(clist.active, False)
        for i, c in enumerate(clist):
            self.assertEqual(c.active, False)
        self.assertNotEqual(
            len(list(clist.components())), len(list(clist.components(active=None)))
        )
        self.assertEqual(len(list(clist.components(active=True))), 0)

        clist[-1].activate()

        self.assertEqual(m.active, False)
        self.assertEqual(blist.active, False)
        self.assertEqual(blist[1].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(clist.active, False)
        clist.activate()
        self.assertEqual(m.active, False)
        self.assertEqual(blist.active, False)
        self.assertEqual(blist[1].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(clist.active, True)
        for i, c in enumerate(clist):
            if i == len(clist) - 1:
                self.assertEqual(c.active, True)
            else:
                self.assertEqual(c.active, False)
        for i, c in enumerate(clist.components(active=None)):
            if i == len(clist) - 1:
                self.assertEqual(c.active, True)
            else:
                self.assertEqual(c.active, False)
        for c in clist.components():
            self.assertEqual(c.active, True)
        for c in clist.components(active=True):
            self.assertEqual(c.active, True)
        self.assertNotEqual(
            len(list(clist.components())), len(list(clist.components(active=None)))
        )
        self.assertEqual(len(list(clist.components(active=True))), 1)

        clist.deactivate(shallow=False)
        clist.activate(shallow=False)

        self.assertEqual(m.active, False)
        self.assertEqual(blist.active, False)
        self.assertEqual(blist[1].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(clist.active, True)
        for i, c in enumerate(clist):
            self.assertEqual(c.active, True)
        for c in clist.components():
            self.assertEqual(c.active, True)
        for c in clist.components(active=True):
            self.assertEqual(c.active, True)
        self.assertEqual(len(list(clist.components())), len(clist))
        self.assertEqual(
            len(list(clist.components())), len(list(clist.components(active=True)))
        )

    def test_preorder_traversal(self):
        clist, traversal = super(
            _TestActiveListContainerBase, self
        ).test_preorder_traversal()

        descend = lambda x: not x._is_heterogeneous_container

        clist[1].deactivate()
        self.assertEqual(
            [None, '[0]', '[2]'],
            [
                c.name
                for c in pmo.preorder_traversal(clist, active=True, descend=descend)
            ],
        )
        self.assertEqual(
            [id(clist), id(clist[0]), id(clist[2])],
            [
                id(c)
                for c in pmo.preorder_traversal(clist, active=True, descend=descend)
            ],
        )

        clist[1].deactivate(shallow=False)
        self.assertEqual(
            [c.name for c in traversal if c.active],
            [
                c.name
                for c in pmo.preorder_traversal(clist, active=True, descend=descend)
            ],
        )
        self.assertEqual(
            [id(c) for c in traversal if c.active],
            [
                id(c)
                for c in pmo.preorder_traversal(clist, active=True, descend=descend)
            ],
        )

        clist.deactivate()
        self.assertEqual(len(list(pmo.preorder_traversal(clist, active=True))), 0)
        self.assertEqual(len(list(pmo.generate_names(clist, active=True))), 0)

    def test_preorder_traversal_descend_check(self):
        clist, traversal = super(
            _TestActiveListContainerBase, self
        ).test_preorder_traversal_descend_check()

        clist[1].deactivate()

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return not x._is_heterogeneous_container

        descend.seen = []
        order = list(pmo.preorder_traversal(clist, active=True, descend=descend))
        self.assertEqual([None, '[0]', '[2]'], [c.name for c in order])
        self.assertEqual(
            [id(clist), id(clist[0]), id(clist[2])], [id(c) for c in order]
        )
        if clist.ctype._is_heterogeneous_container:
            self.assertEqual([None, '[0]', '[2]'], [c.name for c in descend.seen])
            self.assertEqual(
                [id(clist), id(clist[0]), id(clist[2])], [id(c) for c in descend.seen]
            )
        else:
            self.assertEqual([None], [c.name for c in descend.seen])
            self.assertEqual([id(clist)], [id(c) for c in descend.seen])

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return x.active and (not x._is_heterogeneous_container)

        descend.seen = []
        order = list(pmo.preorder_traversal(clist, active=None, descend=descend))
        self.assertEqual([None, '[0]', '[1]', '[2]'], [c.name for c in order])
        self.assertEqual(
            [id(clist), id(clist[0]), id(clist[1]), id(clist[2])],
            [id(c) for c in order],
        )
        if clist.ctype._is_heterogeneous_container:
            self.assertEqual(
                [None, '[0]', '[1]', '[2]'], [c.name for c in descend.seen]
            )
            self.assertEqual(
                [id(clist), id(clist[0]), id(clist[1]), id(clist[2])],
                [id(c) for c in descend.seen],
            )
        else:
            self.assertEqual([None, '[1]'], [c.name for c in descend.seen])
            self.assertEqual([id(clist), id(clist[1])], [id(c) for c in descend.seen])

        clist[1].deactivate(shallow=False)

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return not x._is_heterogeneous_container

        descend.seen = []
        order = list(pmo.preorder_traversal(clist, active=True, descend=descend))
        self.assertEqual(
            [c.name for c in traversal if c.active], [c.name for c in order]
        )
        self.assertEqual([id(c) for c in traversal if c.active], [id(c) for c in order])
        self.assertEqual(
            [c.name for c in traversal if c.active and c._is_container],
            [c.name for c in descend.seen],
        )
        self.assertEqual(
            [id(c) for c in traversal if c.active and c._is_container],
            [id(c) for c in descend.seen],
        )

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return x.active and (not x._is_heterogeneous_container)

        descend.seen = []
        order = list(pmo.preorder_traversal(clist, active=None, descend=descend))
        self.assertEqual([None, '[0]', '[1]', '[2]'], [c.name for c in order])
        self.assertEqual(
            [id(clist), id(clist[0]), id(clist[1]), id(clist[2])],
            [id(c) for c in order],
        )
        if clist.ctype._is_heterogeneous_container:
            self.assertEqual(
                [None, '[0]', '[1]', '[2]'], [c.name for c in descend.seen]
            )
            self.assertEqual(
                [id(clist), id(clist[0]), id(clist[1]), id(clist[2])],
                [id(c) for c in descend.seen],
            )
        else:
            self.assertEqual([None, '[1]'], [c.name for c in descend.seen])
            self.assertEqual([id(clist), id(clist[1])], [id(c) for c in descend.seen])

        clist.deactivate()

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return True

        descend.seen = []
        order = list(pmo.preorder_traversal(clist, active=True, descend=descend))
        self.assertEqual(len(descend.seen), 0)
        self.assertEqual(len(list(pmo.generate_names(clist, active=True))), 0)

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return x.active

        descend.seen = []
        order = list(pmo.preorder_traversal(clist, active=None, descend=descend))
        self.assertEqual(len(descend.seen), 1)
        self.assertIs(descend.seen[0], clist)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for expression generation
#
import logging
import math

import pyomo.common.unittest as unittest

from pyomo.common.log import LoggingIntercept
from pyomo.core.expr.numvalue import is_fixed
from pyomo.core.expr.compare import assertExpressionsStructurallyEqual
from pyomo.core.expr import (
    value,
    sin,
    Expr_if,
    SumExpression,
    LinearExpression,
    NPV_SumExpression,
    AbsExpression,
    DivisionExpression,
    NPV_DivisionExpression,
    ExternalFunctionExpression,
    NPV_ExternalFunctionExpression,
    MonomialTermExpression,
    PowExpression,
    NPV_PowExpression,
    ProductExpression,
    NPV_ProductExpression,
    NegationExpression,
    NPV_NegationExpression,
    UnaryFunctionExpression,
    NPV_UnaryFunctionExpression,
)
from pyomo.core.expr.numeric_expr import (
    mutable_expression,
    nonlinear_expression,
    linear_expression,
    _MutableSumExpression,
    _MutableLinearExpression,
    _MutableNPVSumExpression,
    Expr_ifExpression,
    NPV_Expr_ifExpression,
    MaxExpression,
    NPV_MaxExpression,
    MinExpression,
    NPV_MinExpression,
)
from pyomo.environ import ConcreteModel, Param, Var, ExternalFunction


class MockExternalFunction:
    def evaluate(self, args):
        (x,) = args
        return (math.log(x) / math.log(2)) ** 2

    def getname(self):
        return 'mock_fcn'


class TestExpressionAPI(unittest.TestCase):
    def test_deprecated_functions(self):
        m = ConcreteModel()
        m.x = Var()
        e = m.x**10
        self.assertIs(type(e), PowExpression)
        with LoggingIntercept() as LOG:
            f = e.create_potentially_variable_object()
        self.assertIs(e, f)
        self.assertIs(type(e), PowExpression)
        self.assertIn(
            'DEPRECATED: The implicit recasting of a "not potentially variable" '
            'expression node to a potentially variable one is no longer supported',
            LOG.getvalue().replace('\n', ' '),
        )
        self.assertNotIn(
            'recasting a non-potentially variable expression to a potentially variable '
            'one violates the immutability promise for Pyomo expression trees.',
            LOG.getvalue().replace('\n', ' '),
        )

        m.p = Param(mutable=True)
        e = m.p**10
        self.assertIs(type(e), NPV_PowExpression)
        with LoggingIntercept() as LOG:
            f = e.create_potentially_variable_object()
        self.assertIs(e, f)
        self.assertIs(type(e), PowExpression)
        self.assertIn(
            'DEPRECATED: The implicit recasting of a "not potentially variable" '
            'expression node to a potentially variable one is no longer supported',
            LOG.getvalue().replace('\n', ' '),
        )
        self.assertIn(
            'recasting a non-potentially variable expression to a potentially variable '
            'one violates the immutability promise for Pyomo expression trees.',
            LOG.getvalue().replace('\n', ' '),
        )

        e = m.x + m.x
        with LoggingIntercept() as LOG:
            f = e.add(5)
        self.assertIn(
            'DEPRECATED: SumExpression.add() is deprecated.  Please use regular '
            'Python operators',
            LOG.getvalue().replace('\n', ' '),
        )
        self.assertEqual(str(e), 'x + x')
        self.assertEqual(str(f), 'x + x + 5')

    def test_mutable_expression(self):
        m = ConcreteModel()
        m.x = Var(range(3))
        with mutable_expression() as e:
            f = e
            self.assertIs(type(e), _MutableNPVSumExpression)
            e += 1
            self.assertIs(e, f)
            self.assertIs(type(e), _MutableNPVSumExpression)
            e += m.x[0]
            self.assertIs(e, f)
            self.assertIs(type(e), _MutableLinearExpression)
            e += 100 * m.x[1]
            self.assertIs(e, f)
            self.assertIs(type(e), _MutableLinearExpression)
            e += m.x[0] ** 2
            self.assertIs(e, f)
            self.assertIs(type(e), _MutableSumExpression)
        self.assertIs(e, f)
        self.assertIs(type(e), SumExpression)

    def test_linear_expression(self):
        m = ConcreteModel()
        m.x = Var(range(3))
        with linear_expression() as e:
            f = e
            self.assertIs(type(e), _MutableNPVSumExpression)
            e += 1
            self.assertIs(e, f)
            self.assertIs(type(e), _MutableNPVSumExpression)
            e += m.x[0]
            self.assertIs(e, f)
            self.assertIs(type(e), _MutableLinearExpression)
            e += 100 * m.x[1]
            self.assertIs(e, f)
            self.assertIs(type(e), _MutableLinearExpression)
            e += m.x[0] ** 2
            self.assertIs(e, f)
            self.assertIs(type(e), _MutableSumExpression)
        self.assertIs(e, f)
        self.assertIs(type(e), SumExpression)

    def test_nonlinear_expression(self):
        m = ConcreteModel()
        m.x = Var(range(3))
        with nonlinear_expression() as e:
            f = e
            self.assertIs(type(e), _MutableSumExpression)
            e += 1
            self.assertIs(e, f)
            self.assertIs(type(e), _MutableSumExpression)
            e += m.x[0]
            self.assertIs(e, f)
            self.assertIs(type(e), _MutableSumExpression)
            e += 100 * m.x[1]
            self.assertIs(e, f)
            self.assertIs(type(e), _MutableSumExpression)
            e += m.x[0] ** 2
            self.assertIs(e, f)
            self.assertIs(type(e), _MutableSumExpression)
        self.assertIs(e, f)
        self.assertIs(type(e), SumExpression)

    def test_negation(self):
        m = ConcreteModel()
        m.x = Var(initialize=5)

        e = NegationExpression((5,))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(value(e), -5)
        self.assertEqual(str(e), "- 5")
        self.assertEqual(e.to_string(verbose=True), "neg(5)")

        e = NegationExpression((-5,))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(value(e), 5)
        self.assertEqual(str(e), "5")
        self.assertEqual(e.to_string(verbose=True), "neg(-5)")

        e = NegationExpression((m.x,))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), -5)
        self.assertEqual(str(e), "- x")
        self.assertEqual(e.to_string(verbose=True), "neg(x)")

        m.p = Param(initialize=10, mutable=True)
        e = NPV_NegationExpression((m.p,))
        self.assertFalse(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(value(e), -10)
        self.assertEqual(str(e), "- p")
        self.assertEqual(e.to_string(verbose=True), "neg(p)")

        e = -(m.x + 2 * m.x)
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), -15)
        self.assertEqual(str(e), "- (x + 2*x)")
        self.assertEqual(e.to_string(verbose=True), "neg(sum(x, mon(2, x)))")

        # This can't occur through operator overloading, but could
        # through expression substitution
        e = NegationExpression((NegationExpression((m.x,)),))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 5)
        self.assertEqual(str(e), "x")
        self.assertEqual(e.to_string(verbose=True), "neg(neg(x))")

    def test_pow(self):
        m = ConcreteModel()
        m.x = Var(initialize=5)
        e = PowExpression((m.x, 2))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 2)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 25)
        self.assertEqual(str(e), "x**2")
        self.assertEqual(e.to_string(verbose=True), "pow(x, 2)")

        e = PowExpression((2, m.x))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), None)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 32)
        self.assertEqual(str(e), "2**x")
        self.assertEqual(e.to_string(verbose=True), "pow(2, x)")

        m.p = Param(initialize=3, mutable=True)
        e = NPV_PowExpression((m.p, 2))
        self.assertFalse(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(value(e), 9)
        self.assertEqual(str(e), "p**2")
        self.assertEqual(e.to_string(verbose=True), "pow(p, 2)")

        e = NPV_PowExpression((2, m.p))
        self.assertFalse(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(value(e), 8)
        self.assertEqual(str(e), "2**p")
        self.assertEqual(e.to_string(verbose=True), "pow(2, p)")

        e = PowExpression((m.x, m.p))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 3)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 125)
        self.assertEqual(str(e), "x**p")
        self.assertEqual(e.to_string(verbose=True), "pow(x, p)")

        m.p = 0
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(value(e), 1)
        self.assertEqual(str(e), "x**p")
        self.assertEqual(e.to_string(verbose=True), "pow(x, p)")

        m.x.fix(2)
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(value(e), 1)
        self.assertEqual(str(e), "x**p")
        self.assertEqual(e.to_string(verbose=True), "pow(x, p)")

        m.p = 3
        e = PowExpression((m.p, m.x))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(value(e), 9)
        self.assertEqual(str(e), "p**x")
        self.assertEqual(e.to_string(verbose=True), "pow(p, x)")

        m.y = Var()
        m.x.fix(None)
        e = PowExpression((m.y, m.x))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), None)
        self.assertEqual(is_fixed(e), False)
        with self.assertRaisesRegex(
            ValueError, 'No value for uninitialized ScalarVar object y'
        ):
            self.assertEqual(value(e), None)
        self.assertEqual(str(e), "y**x")
        self.assertEqual(e.to_string(verbose=True), "pow(y, x)")

    def test_min(self):
        m = ConcreteModel()
        m.x = Var(initialize=5)
        e = MinExpression((m.x, 2))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), None)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 2)
        self.assertEqual(str(e), "min(x, 2)")
        self.assertEqual(e.to_string(verbose=True), "min(x, 2)")

        m.x.fix(1)
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(value(e), 1)
        self.assertEqual(str(e), "min(x, 2)")
        self.assertEqual(e.to_string(verbose=True), "min(x, 2)")

    def test_max(self):
        m = ConcreteModel()
        m.x = Var(initialize=5)
        e = MaxExpression((m.x, 2))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), None)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 5)
        self.assertEqual(str(e), "max(x, 2)")
        self.assertEqual(e.to_string(verbose=True), "max(x, 2)")

        m.x.fix(10)
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(value(e), 10)
        self.assertEqual(str(e), "max(x, 2)")
        self.assertEqual(e.to_string(verbose=True), "max(x, 2)")

    def test_prod(self):
        m = ConcreteModel()
        m.x = Var(initialize=5)
        e = ProductExpression((m.x, 2))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 10)
        self.assertEqual(str(e), "x*2")
        self.assertEqual(e.to_string(verbose=True), "prod(x, 2)")

        e = ProductExpression((2, m.x))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 10)
        self.assertEqual(str(e), "2*x")
        self.assertEqual(e.to_string(verbose=True), "prod(2, x)")

        m.p = Param(initialize=3, mutable=True)
        e = NPV_ProductExpression((m.p, 2))
        self.assertFalse(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(value(e), 6)
        self.assertEqual(str(e), "p*2")
        self.assertEqual(e.to_string(verbose=True), "prod(p, 2)")

        e = NPV_ProductExpression((2, m.p))
        self.assertFalse(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(value(e), 6)
        self.assertEqual(str(e), "2*p")
        self.assertEqual(e.to_string(verbose=True), "prod(2, p)")

        e = ProductExpression((m.x, m.p))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 15)
        self.assertEqual(str(e), "x*p")
        self.assertEqual(e.to_string(verbose=True), "prod(x, p)")

        m.p = 0
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(value(e), 0)
        self.assertEqual(str(e), "x*p")
        self.assertEqual(e.to_string(verbose=True), "prod(x, p)")

        e = ProductExpression((m.p, m.x))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(value(e), 0)
        self.assertEqual(str(e), "p*x")
        self.assertEqual(e.to_string(verbose=True), "prod(p, x)")

        e = ProductExpression((-1, m.x))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), -5)
        self.assertEqual(str(e), "- x")
        self.assertEqual(e.to_string(verbose=True), "prod(-1, x)")

        m.y = Var()
        e = ProductExpression((m.y, m.x))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 2)
        self.assertEqual(is_fixed(e), False)
        with self.assertRaisesRegex(
            ValueError, 'No value for uninitialized ScalarVar object y'
        ):
            self.assertEqual(value(e), None)
        self.assertEqual(str(e), "y*x")
        self.assertEqual(e.to_string(verbose=True), "prod(y, x)")

        m.x.fix(0)
        e = ProductExpression((m.y, m.x))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        with self.assertRaisesRegex(
            ValueError, 'No value for uninitialized ScalarVar object y'
        ):
            self.assertEqual(value(e), None)
        self.assertEqual(str(e), "y*x")
        self.assertEqual(e.to_string(verbose=True), "prod(y, x)")

        m.x.fix(None)
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        with self.assertRaisesRegex(
            ValueError, 'No value for uninitialized ScalarVar object y'
        ):
            self.assertEqual(value(e), None)
        self.assertEqual(str(e), "y*x")
        self.assertEqual(e.to_string(verbose=True), "prod(y, x)")

        m.y = 5
        e = ProductExpression((1 / m.y, m.x))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), None)
        self.assertEqual(is_fixed(e), False)
        with self.assertRaisesRegex(
            ValueError, 'No value for uninitialized ScalarVar object x'
        ):
            self.assertEqual(value(e), None)
        self.assertEqual(str(e), "1/y*x")
        self.assertEqual(e.to_string(verbose=True), "prod(div(1, y), x)")

    def test_monomial(self):
        m = ConcreteModel()
        m.x = Var(initialize=5)
        e = MonomialTermExpression((2, m.x))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 10)
        self.assertEqual(str(e), "2*x")
        self.assertEqual(e.to_string(verbose=True), "mon(2, x)")

        e = MonomialTermExpression((-2, m.x))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), -10)
        self.assertEqual(str(e), "-2*x")
        self.assertEqual(e.to_string(verbose=True), "mon(-2, x)")

        m.x.fix(2)
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(value(e), -4)
        self.assertEqual(str(e), "-2*x")
        self.assertEqual(e.to_string(verbose=True), "mon(-2, x)")

    def test_division(self):
        m = ConcreteModel()
        m.x = Var(initialize=5)

        e = DivisionExpression((m.x, 2))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 2.5)
        self.assertEqual(str(e), "x/2")
        self.assertEqual(e.to_string(verbose=True), "div(x, 2)")

        e = DivisionExpression((2, m.x))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), None)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 0.4)
        self.assertEqual(str(e), "2/x")
        self.assertEqual(e.to_string(verbose=True), "div(2, x)")

        m.x.fix(2)
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(value(e), 1)
        self.assertEqual(str(e), "2/x")
        self.assertEqual(e.to_string(verbose=True), "div(2, x)")

    def test_sum(self):
        m = ConcreteModel()
        m.x = Var(initialize=5)

        e = SumExpression(())
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(value(e), 0)
        self.assertEqual(str(e), "0")
        self.assertEqual(e.to_string(verbose=True), "sum(0)")

        e = SumExpression((m.x, 2))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 7)
        self.assertEqual(str(e), "x + 2")
        self.assertEqual(e.to_string(verbose=True), "sum(x, 2)")

        e = SumExpression((m.x, -2))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 3)
        self.assertEqual(str(e), "x - 2")
        self.assertEqual(e.to_string(verbose=True), "sum(x, -2)")

        e = SumExpression((-2, m.x, -2))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 1)
        self.assertEqual(str(e), "-2 + x - 2")
        self.assertEqual(e.to_string(verbose=True), "sum(-2, x, -2)")

        e = SumExpression([-2, m.x, AbsExpression((m.x,)), m.x])
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), None)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 13)
        self.assertEqual(str(e), "-2 + x + abs(x) + x")
        self.assertEqual(e.to_string(verbose=True), "sum(-2, x, abs(x), x)")

        e = SumExpression([-2, m.x, SumExpression([-2])])
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 1)
        self.assertEqual(str(e), "-2 + x + (-2)")
        self.assertEqual(e.to_string(verbose=True), "sum(-2, x, sum(-2))")

        e = SumExpression([-2, m.x, SumExpression([-2, 3])])
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 4)
        self.assertEqual(str(e), "-2 + x + (-2 + 3)")
        self.assertEqual(e.to_string(verbose=True), "sum(-2, x, sum(-2, 3))")

        e = SumExpression([-2, m.x, SumExpression([2, 3])])
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 8)
        self.assertEqual(str(e), "-2 + x + (2 + 3)")
        self.assertEqual(e.to_string(verbose=True), "sum(-2, x, sum(2, 3))")

        e = SumExpression([-2, m.x, NegationExpression((SumExpression([-2]),))])
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 5)
        self.assertEqual(str(e), "-2 + x - (-2)")
        self.assertEqual(e.to_string(verbose=True), "sum(-2, x, neg(sum(-2)))")

        e = SumExpression((NegationExpression((m.x,)), -2))
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), -7)
        self.assertEqual(str(e), "- x - 2")
        self.assertEqual(e.to_string(verbose=True), "sum(neg(x), -2)")

        m.x.fix(2)
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(value(e), -4)
        self.assertEqual(str(e), "- x - 2")
        self.assertEqual(e.to_string(verbose=True), "sum(neg(x), -2)")

    def test_linear(self):
        m = ConcreteModel()
        m.x = Var(range(3), initialize=range(3))
        m.y = Var(initialize=5)
        with mutable_expression() as e:
            for i in range(3):
                e += i * m.x[i]
            e += 5
            e += m.y
            e -= 3

        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 1 + 4 + 5 + 2)
        self.assertEqual(str(e), "0*x[0] + x[1] + 2*x[2] + 5 + y - 3")
        self.assertEqual(
            e.to_string(verbose=True), "sum(mon(0, x[0]), x[1], mon(2, x[2]), 5, y, -3)"
        )

        self.assertIs(type(e), LinearExpression)
        self.assertEqual(e.constant, 2)
        cache = e._cache
        self.assertEqual(e.linear_coefs, (0, 1, 2, 1))
        self.assertIs(cache, e._cache)
        self.assertEqual(e.linear_vars, (m.x[0], m.x[1], m.x[2], m.y))
        self.assertIs(cache, e._cache)

        e = LinearExpression()
        self.assertEqual(e.linear_coefs, tuple())
        self.assertIsNot(cache, e._cache)
        cache = e._cache
        e = LinearExpression()
        self.assertEqual(e.constant, 0)
        self.assertIsNot(cache, e._cache)
        cache = e._cache
        e = LinearExpression()
        self.assertEqual(e.linear_vars, tuple())
        self.assertIsNot(cache, e._cache)

        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(value(e), 0)
        self.assertEqual(str(e), "0")
        self.assertEqual(e.to_string(verbose=True), "sum(0)")

        e = LinearExpression(constant=5, linear_vars=[m.y, m.x[1]], linear_coefs=[3, 5])
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 25)
        self.assertEqual(str(e), "5 + 3*y + 5*x[1]")
        self.assertEqual(e.to_string(verbose=True), "sum(5, mon(3, y), mon(5, x[1]))")

        with self.assertRaisesRegex(
            ValueError,
            "Cannot specify both args and any of "
            "{constant, linear_coefs, or linear_vars}",
        ):
            LinearExpression(5, constant=5)

        with self.assertRaisesRegex(
            ValueError,
            r"linear_vars \(\[y\]\) is not compatible with linear_coefs \(\[3, 5\]\)",
        ):
            LinearExpression(constant=5, linear_vars=[m.y], linear_coefs=[3, 5])

    def test_expr_if(self):
        m = ConcreteModel()
        m.x = Var(range(3), initialize=range(3))
        m.y = Var(initialize=5)
        e = Expr_if(IF=m.y >= 5, THEN=m.x[0] + 5, ELSE=m.x[1] ** 2)

        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), None)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 5)
        self.assertEqual(
            str(e), "Expr_if( ( 5  <=  y ), then=( x[0] + 5 ), else=( x[1]**2 ) )"
        )
        self.assertEqual(
            e.to_string(verbose=True),
            "Expr_if( ( 5  <=  y ), then=( sum(x[0], 5) ), else=( pow(x[1], 2) ) )",
        )

        m.y.fix()
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 5)

        m.y.fix(4)
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 2)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 1)

        e = Expr_if(IF=m.y >= 5, THEN=m.x[0] + 5, ELSE=m.x[1] + 10)
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 11)

        m.y.fix(5)
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 5)

    def test_unary(self):
        m = ConcreteModel()
        m.x = Var(initialize=5)
        e = sin(2 * m.x)
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), None)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), math.sin(10))
        self.assertEqual(str(e), "sin(2*x)")
        self.assertEqual(e.to_string(verbose=True), "sin(mon(2, x))")

        m.x.fix(1)
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(value(e), math.sin(2))
        self.assertEqual(str(e), "sin(2*x)")
        self.assertEqual(e.to_string(verbose=True), "sin(mon(2, x))")

    def test_external(self):
        m = ConcreteModel()
        m.x = Var(initialize=16)
        fcn = MockExternalFunction()
        e = ExternalFunctionExpression((2 * m.x,), fcn)
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), None)
        self.assertEqual(is_fixed(e), False)
        self.assertEqual(value(e), 25)
        self.assertEqual(str(e), "mock_fcn(2*x)")
        self.assertEqual(e.to_string(verbose=True), "mock_fcn(mon(2, x))")

        m.x.fix(1)
        self.assertTrue(e.is_potentially_variable())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(value(e), 1)
        self.assertEqual(str(e), "mock_fcn(2*x)")
        self.assertEqual(e.to_string(verbose=True), "mock_fcn(mon(2, x))")


class TestExpressionDuplicateAPI(unittest.TestCase):
    def test_negation(self):
        m = ConcreteModel()
        m.x = Var(initialize=5)
        m.p = Param(initialize=3, mutable=True)
        e = NegationExpression((m.x,))

        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertIs(f.args, e.args)

        f = e.create_node_with_local_data((2,))
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertEqual(f.args, (2,))

        e = NPV_NegationExpression((m.p,))
        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))

        f = e.create_node_with_local_data((m.x,))
        self.assertIsNot(f, e)
        self.assertIs(type(f), NegationExpression)

    def test_pow(self):
        m = ConcreteModel()
        m.x = Var(initialize=5)
        m.p = Param(initialize=3, mutable=True)
        e = PowExpression((m.x, 2))

        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertIs(f.args, e.args)

        f = e.create_node_with_local_data((2, 3))
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertEqual(f.args, (2, 3))

        e = NPV_PowExpression((m.p, 2))
        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))

        f = e.create_node_with_local_data((m.p, m.x))
        self.assertIsNot(f, e)
        self.assertIs(type(f), PowExpression)

    def test_min(self):
        m = ConcreteModel()
        m.x = Var(initialize=5)
        m.p = Param(initialize=3, mutable=True)
        e = MinExpression((m.x, 2))

        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertIs(f.args, e.args)

        f = e.create_node_with_local_data((2, 3, 4))
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertEqual(f.args, (2, 3, 4))

        e = NPV_MinExpression((m.p, 2))
        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))

        f = e.create_node_with_local_data((m.p, m.x))
        self.assertIsNot(f, e)
        self.assertIs(type(f), MinExpression)

    def test_max(self):
        m = ConcreteModel()
        m.x = Var(initialize=5)
        m.p = Param(initialize=3, mutable=True)
        e = MaxExpression((m.x, 2))

        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertIs(f.args, e.args)

        f = e.create_node_with_local_data((2, 3, 4))
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertEqual(f.args, (2, 3, 4))

        e = NPV_MaxExpression((m.p, 2))
        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))

        f = e.create_node_with_local_data((m.p, m.x))
        self.assertIsNot(f, e)
        self.assertIs(type(f), MaxExpression)

    def test_prod(self):
        m = ConcreteModel()
        m.x = Var(initialize=5)
        m.p = Param(initialize=3, mutable=True)
        e = ProductExpression((m.x, 2))

        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertIs(f.args, e.args)

        f = e.create_node_with_local_data((m.p, 3))
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertEqual(f.args, (m.p, 3))

        e = NPV_ProductExpression((m.p, 2))
        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))

        f = e.create_node_with_local_data((m.p, m.x))
        self.assertIsNot(f, e)
        self.assertIs(type(f), ProductExpression)

    def test_monomial(self):
        m = ConcreteModel()
        m.x = Var(initialize=5)
        m.p = Param(initialize=3, mutable=True)
        e = MonomialTermExpression((2, m.x))

        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        # Note that MonomialTermExpression recreates the args tuple
        self.assertEqual(f.args, e.args)

        f = e.create_node_with_local_data((m.p, 3))
        self.assertIsNot(f, e)
        self.assertIs(type(f), NPV_ProductExpression)
        self.assertEqual(f.args, (m.p, 3))

        f = e.create_node_with_local_data((m.x, m.x))
        self.assertIsNot(f, e)
        self.assertIs(type(f), ProductExpression)
        self.assertEqual(f.args, (m.x, m.x))

    def test_division(self):
        m = ConcreteModel()
        m.x = Var(initialize=5)
        m.p = Param(initialize=3, mutable=True)
        e = DivisionExpression((m.x, 2))

        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertIs(f.args, e.args)

        f = e.create_node_with_local_data((2, 3))
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertEqual(f.args, (2, 3))

        e = NPV_DivisionExpression((m.p, 2))
        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))

        f = e.create_node_with_local_data((m.p, m.x))
        self.assertIsNot(f, e)
        self.assertIs(type(f), DivisionExpression)

    def test_sum(self):
        m = ConcreteModel()
        m.x = Var(initialize=5)
        m.p = Param(initialize=3, mutable=True)
        e = SumExpression((m.x, 2))

        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertIsNot(f._args_, e._args_)
        self.assertIsNot(f.args, e.args)

        f = e.create_node_with_local_data(e._args_)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertIs(f._args_, e._args_)
        self.assertIsNot(f.args, e.args)

        f = e.create_node_with_local_data((m.x, 2, 3))
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertEqual(f.args, [m.x, 2, 3])

        e = NPV_SumExpression((m.p, 2))
        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))

        f = e.create_node_with_local_data((m.p, m.x))
        self.assertIsNot(f, e)
        self.assertIs(type(f), LinearExpression)
        assertExpressionsStructurallyEqual(self, f.args, [m.p, m.x])

        f = e.create_node_with_local_data((m.p, m.x**2))
        self.assertIsNot(f, e)
        self.assertIs(type(f), SumExpression)
        assertExpressionsStructurallyEqual(self, f.args, [m.p, PowExpression((m.x, 2))])

    def test_linear(self):
        m = ConcreteModel()
        m.x = Var(range(3), initialize=range(3))
        m.y = Var(initialize=5)
        with mutable_expression() as e:
            for i in range(3):
                e += i * m.x[i]
            e += 5
            e += m.y
            e -= 3

    def test_expr_if(self):
        m = ConcreteModel()
        m.x = Var(range(3), initialize=range(3))
        m.y = Var(initialize=5)
        m.p = Param(initialize=3, mutable=True)
        e = Expr_if(IF=m.y >= 5, THEN=m.x[0] + 5, ELSE=m.x[1] ** 2)

        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertIs(f.args, e.args)

        f = e.create_node_with_local_data((2, 3, 4))
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertEqual(f.args, (2, 3, 4))

        e = NPV_Expr_ifExpression((m.p <= 5, 2, m.p))
        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))

        f = e.create_node_with_local_data((m.p <= 5, m.x, m.p))
        self.assertIsNot(f, e)
        self.assertIs(type(f), Expr_ifExpression)

    def test_unary_fcn(self):
        m = ConcreteModel()
        m.x = Var(range(3), initialize=range(3))
        m.y = Var(initialize=5)
        m.p = Param(initialize=3, mutable=True)
        e = sin(2 * m.y)

        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertIs(f.args, e.args)
        self.assertIs(e._fcn, f._fcn)
        self.assertIs(e._name, f._name)

        f = e.create_node_with_local_data((m.x[1],))
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertEqual(f.args, (m.x[1],))
        self.assertIs(e._fcn, f._fcn)
        self.assertIs(e._name, f._name)

        e = sin((2 * m.p))
        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertIs(e._fcn, f._fcn)
        self.assertIs(e._name, f._name)

        f = e.create_node_with_local_data((m.x[1],))
        self.assertIsNot(f, e)
        self.assertIs(type(f), UnaryFunctionExpression)
        self.assertEqual(f.args, (m.x[1],))
        self.assertIs(e._fcn, f._fcn)
        self.assertIs(e._name, f._name)

    def test_abs(self):
        m = ConcreteModel()
        m.x = Var(range(3), initialize=range(3))
        m.y = Var(initialize=5)
        m.p = Param(initialize=3, mutable=True)
        e = abs(2 * m.y)

        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertIs(f.args, e.args)
        self.assertIs(e._fcn, f._fcn)
        self.assertIs(e._name, f._name)

        f = e.create_node_with_local_data((m.x[1],))
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertEqual(f.args, (m.x[1],))
        self.assertIs(e._fcn, f._fcn)
        self.assertIs(e._name, f._name)

        e = abs((2 * m.p))
        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertIs(e._fcn, f._fcn)
        self.assertIs(e._name, f._name)

        f = e.create_node_with_local_data((m.x[1],))
        self.assertIsNot(f, e)
        self.assertIs(type(f), AbsExpression)
        self.assertEqual(f.args, (m.x[1],))
        self.assertIs(e._fcn, f._fcn)
        self.assertIs(e._name, f._name)

    def test_external(self):
        m = ConcreteModel()
        m.x = Var(initialize=16)
        m.p = Param(initialize=32, mutable=True)
        fcn = MockExternalFunction()
        e = ExternalFunctionExpression((2 * m.x,), fcn)

        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertIs(f.args, e.args)
        self.assertIs(e._fcn, f._fcn)

        f = e.create_node_with_local_data((m.x,))
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertEqual(f.args, (m.x,))
        self.assertIs(e._fcn, f._fcn)

        e = NPV_ExternalFunctionExpression((2 * m.p,), fcn)
        f = e.create_node_with_local_data(e.args)
        self.assertIsNot(f, e)
        self.assertIs(type(f), type(e))
        self.assertIs(e._fcn, f._fcn)

        f = e.create_node_with_local_data((m.x,))
        self.assertIsNot(f, e)
        self.assertIs(type(f), ExternalFunctionExpression)
        self.assertEqual(f.args, (m.x,))
        self.assertIs(e._fcn, f._fcn)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from io import StringIO
from typing import Sequence, Dict, Optional, Mapping, MutableMapping

import pyomo.environ as pyo
from pyomo.common.config import ConfigDict
from pyomo.core.base.constraint import ConstraintData
from pyomo.core.base.var import VarData
from pyomo.common.collections import ComponentMap
from pyomo.contrib.solver.common import results
from pyomo.contrib.solver.common import solution_loader
from pyomo.core.base.var import Var
from pyomo.common import unittest


class SolutionLoaderExample(solution_loader.SolutionLoaderBase):
    """
    This is an example instantiation of a SolutionLoader that is used for
    testing generated results.
    """

    def __init__(
        self,
        primals: Optional[MutableMapping],
        duals: Optional[MutableMapping],
        reduced_costs: Optional[MutableMapping],
    ):
        """
        Parameters
        ----------
        primals: dict
            maps id(Var) to (var, value)
        duals: dict
            maps Constraint to dual value
        reduced_costs: dict
            maps id(Var) to (var, reduced_cost)
        """
        self._primals = primals
        self._duals = duals
        self._reduced_costs = reduced_costs

    def get_primals(
        self, vars_to_load: Optional[Sequence[VarData]] = None
    ) -> Mapping[VarData, float]:
        if self._primals is None:
            raise RuntimeError(
                'Solution loader does not currently have a valid solution. Please '
                'check the termination condition.'
            )
        if vars_to_load is None:
            return ComponentMap(self._primals.values())
        else:
            primals = ComponentMap()
            for v in vars_to_load:
                primals[v] = self._primals[id(v)][1]
            return primals

    def get_duals(
        self, cons_to_load: Optional[Sequence[ConstraintData]] = None
    ) -> Dict[ConstraintData, float]:
        if self._duals is None:
            raise RuntimeError(
                'Solution loader does not currently have valid duals. Please '
                'check the termination condition and ensure the solver returns duals '
                'for the given problem type.'
            )
        if cons_to_load is None:
            duals = dict(self._duals)
        else:
            duals = {}
            for c in cons_to_load:
                duals[c] = self._duals[c]
        return duals

    def get_reduced_costs(
        self, vars_to_load: Optional[Sequence[VarData]] = None
    ) -> Mapping[VarData, float]:
        if self._reduced_costs is None:
            raise RuntimeError(
                'Solution loader does not currently have valid reduced costs. Please '
                'check the termination condition and ensure the solver returns reduced '
                'costs for the given problem type.'
            )
        if vars_to_load is None:
            rc = ComponentMap(self._reduced_costs.values())
        else:
            rc = ComponentMap()
            for v in vars_to_load:
                rc[v] = self._reduced_costs[id(v)][1]
        return rc


class TestTerminationCondition(unittest.TestCase):
    def test_member_list(self):
        member_list = results.TerminationCondition._member_names_
        expected_list = [
            'unknown',
            'convergenceCriteriaSatisfied',
            'maxTimeLimit',
            'iterationLimit',
            'objectiveLimit',
            'minStepLength',
            'unbounded',
            'provenInfeasible',
            'locallyInfeasible',
            'infeasibleOrUnbounded',
            'error',
            'interrupted',
            'licensingProblems',
        ]
        self.assertEqual(member_list.sort(), expected_list.sort())

    def test_codes(self):
        self.assertEqual(results.TerminationCondition.unknown.value, 42)
        self.assertEqual(
            results.TerminationCondition.convergenceCriteriaSatisfied.value, 0
        )
        self.assertEqual(results.TerminationCondition.maxTimeLimit.value, 1)
        self.assertEqual(results.TerminationCondition.iterationLimit.value, 2)
        self.assertEqual(results.TerminationCondition.objectiveLimit.value, 3)
        self.assertEqual(results.TerminationCondition.minStepLength.value, 4)
        self.assertEqual(results.TerminationCondition.unbounded.value, 5)
        self.assertEqual(results.TerminationCondition.provenInfeasible.value, 6)
        self.assertEqual(results.TerminationCondition.locallyInfeasible.value, 7)
        self.assertEqual(results.TerminationCondition.infeasibleOrUnbounded.value, 8)
        self.assertEqual(results.TerminationCondition.error.value, 9)
        self.assertEqual(results.TerminationCondition.interrupted.value, 10)
        self.assertEqual(results.TerminationCondition.licensingProblems.value, 11)


class TestSolutionStatus(unittest.TestCase):
    def test_member_list(self):
        member_list = results.SolutionStatus._member_names_
        expected_list = ['noSolution', 'unknown', 'infeasible', 'feasible', 'optimal']
        self.assertEqual(member_list, expected_list)

    def test_codes(self):
        self.assertEqual(results.SolutionStatus.noSolution.value, 0)
        self.assertEqual(results.SolutionStatus.infeasible.value, 10)
        self.assertEqual(results.SolutionStatus.feasible.value, 20)
        self.assertEqual(results.SolutionStatus.optimal.value, 30)


class TestResults(unittest.TestCase):
    def test_member_list(self):
        res = results.Results()
        expected_declared = {
            'extra_info',
            'incumbent_objective',
            'objective_bound',
            'solution_loader',
            'solution_status',
            'solver_name',
            'solver_version',
            'termination_condition',
            'timing_info',
            'solver_log',
            'solver_config',
        }
        actual_declared = res._declared
        self.assertEqual(expected_declared, actual_declared)

    def test_default_initialization(self):
        res = results.Results()
        self.assertIsNone(res.solution_loader)
        self.assertIsNone(res.incumbent_objective)
        self.assertIsNone(res.objective_bound)
        self.assertEqual(
            res.termination_condition, results.TerminationCondition.unknown
        )
        self.assertEqual(res.solution_status, results.SolutionStatus.noSolution)
        self.assertIsNone(res.solver_name)
        self.assertIsNone(res.solver_version)
        self.assertIsInstance(res.timing_info, ConfigDict)
        self.assertIsInstance(res.extra_info, ConfigDict)
        self.assertIsNone(res.timing_info.start_timestamp)
        self.assertIsNone(res.timing_info.wall_time)

    def test_display(self):
        res = results.Results()
        stream = StringIO()
        res.display(ostream=stream)
        expected_print = """termination_condition: TerminationCondition.unknown
solution_status: SolutionStatus.noSolution
incumbent_objective: None
objective_bound: None
solver_name: None
solver_version: None
timing_info:
  start_timestamp: None
  wall_time: None
extra_info:
"""
        out = stream.getvalue()
        if 'null' in out:
            out = out.replace('null', 'None')
        self.assertEqual(expected_print, out)

    def test_generated_results(self):
        m = pyo.ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.c1 = pyo.Constraint(expr=m.x == 1)
        m.c2 = pyo.Constraint(expr=m.y == 2)

        primals = {}
        primals[id(m.x)] = (m.x, 1)
        primals[id(m.y)] = (m.y, 2)
        duals = {}
        duals[m.c1] = 3
        duals[m.c2] = 4
        rc = {}
        rc[id(m.x)] = (m.x, 5)
        rc[id(m.y)] = (m.y, 6)

        res = results.Results()
        res.solution_loader = SolutionLoaderExample(
            primals=primals, duals=duals, reduced_costs=rc
        )

        res.solution_loader.load_vars()
        self.assertAlmostEqual(m.x.value, 1)
        self.assertAlmostEqual(m.y.value, 2)

        m.x.value = None
        m.y.value = None

        res.solution_loader.load_vars([m.y])
        self.assertIsNone(m.x.value)
        self.assertAlmostEqual(m.y.value, 2)

        duals2 = res.solution_loader.get_duals()
        self.assertAlmostEqual(duals[m.c1], duals2[m.c1])
        self.assertAlmostEqual(duals[m.c2], duals2[m.c2])

        duals2 = res.solution_loader.get_duals([m.c2])
        self.assertNotIn(m.c1, duals2)
        self.assertAlmostEqual(duals[m.c2], duals2[m.c2])

        rc2 = res.solution_loader.get_reduced_costs()
        self.assertAlmostEqual(rc[id(m.x)][1], rc2[m.x])
        self.assertAlmostEqual(rc[id(m.y)][1], rc2[m.y])

        rc2 = res.solution_loader.get_reduced_costs([m.y])
        self.assertNotIn(m.x, rc2)
        self.assertAlmostEqual(rc[id(m.y)][1], rc2[m.y])



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for BuildAction() Objects
#
# PyomoModel                Base test class
# Scalar                    Test scalar parameter
# Array1                    Test arrays of parameters
#

from io import StringIO
import os

import pyomo.common.unittest as unittest

from pyomo.environ import AbstractModel, Param, Set, BuildAction, value


def action1_fn(model):
    model.A = 4.3


def action2_fn(model, i):
    if i in model.A:
        model.A[i] = value(model.A[i]) + i


def action3_fn(model, i):
    if i in model.A.sparse_keys():
        model.A[i] = value(model.A[i]) + i


class Scalar(unittest.TestCase):
    def setUp(self):
        #
        # Create model instance
        #
        model = AbstractModel()
        model.A = Param(initialize=3.3, mutable=True)
        model.action1 = BuildAction(rule=action1_fn)
        self.instance = model.create_instance()

    def tearDown(self):
        self.instance = None
        if os.path.exists("param.dat"):
            os.remove("param.dat")

    def test_value(self):
        """Check the value of the parameter"""
        tmp = value(self.instance.A.value)
        self.assertEqual(type(tmp), float)
        self.assertEqual(tmp, 4.3)
        self.assertEqual(value(self.instance.A.value), value(self.instance.A))

    def test_getattr(self):
        """Check the use of the __getattr__ method"""
        self.assertEqual(self.instance.A.value, 4.3)


class Array_Param(unittest.TestCase):
    def test_sparse_param_nodefault(self):
        #
        # Create model instance
        #
        model = AbstractModel()
        model.Z = Set(initialize=[1, 3])
        model.A = Param(model.Z, initialize={1: 1.3}, mutable=True)
        model.action2 = BuildAction(model.Z, rule=action2_fn)
        instance = model.create_instance()

        tmp = value(instance.A[1])
        self.assertEqual(type(tmp), float)
        self.assertEqual(tmp, 2.3)

    def test_sparse_param_nodefault_sparse_iter(self):
        #
        # Create model instance
        #
        model = AbstractModel()
        model.Z = Set(initialize=[1, 3])
        model.A = Param(model.Z, initialize={1: 1.3}, mutable=True)
        model.action2 = BuildAction(model.Z, rule=action3_fn)
        instance = model.create_instance()

        tmp = value(instance.A[1])
        self.assertEqual(type(tmp), float)
        self.assertEqual(tmp, 2.3)

    def test_sparse_param_default(self):
        #
        # Create model instance
        #
        model = AbstractModel()
        model.Z = Set(initialize=[1, 3])
        model.A = Param(model.Z, initialize={1: 1.3}, default=0, mutable=True)
        model.action2 = BuildAction(model.Z, rule=action2_fn)
        instance = model.create_instance()

        tmp = value(instance.A[1])
        self.assertEqual(type(tmp), float)
        self.assertEqual(tmp, 2.3)

    def test_dense_param(self):
        #
        # Create model instance
        #
        model = AbstractModel()
        model.Z = Set(initialize=[1, 3])
        model.A = Param(model.Z, initialize=1.3, mutable=True)
        model.action2 = BuildAction(model.Z, rule=action2_fn)
        instance = model.create_instance()
        #
        self.assertEqual(instance.A[1].value, 2.3)
        self.assertEqual(value(instance.A[3]), 4.3)
        #
        buf = StringIO()
        instance.pprint(ostream=buf)
        self.assertEqual(
            buf.getvalue(),
            """1 Set Declarations
    Z : Size=1, Index=None, Ordered=Insertion
        Key  : Dimen : Domain : Size : Members
        None :     1 :    Any :    2 : {1, 3}

1 Param Declarations
    A : Size=2, Index=Z, Domain=Any, Default=None, Mutable=True
        Key : Value
          1 :   2.3
          3 :   4.3

1 BuildAction Declarations
    action2 : Size=0, Index=Z, Active=True

3 Declarations: Z A action2
""",
        )


class TestMisc(unittest.TestCase):
    def test_error1(self):
        model = AbstractModel()
        try:
            model.a = BuildAction()
            self.fail("Expected ValueError")
        except ValueError:
            pass


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
import pyomo.dae as dae
from pyomo.common.dependencies import networkx_available
from pyomo.common.dependencies import scipy_available
from pyomo.common.collections import ComponentSet, ComponentMap
from pyomo.contrib.incidence_analysis.scc_solver import (
    TemporarySubsystemManager,
    generate_strongly_connected_components,
    solve_strongly_connected_components,
)
from pyomo.contrib.incidence_analysis.tests.models_for_testing import (
    make_gas_expansion_model,
    make_dynamic_model,
)
import pyomo.common.unittest as unittest


@unittest.skipUnless(scipy_available, "SciPy is not available")
@unittest.skipUnless(networkx_available, "NetworkX is not available")
class TestGenerateSCC(unittest.TestCase):
    def test_gas_expansion(self):
        N = 5
        m = make_gas_expansion_model(N)
        m.rho[0].fix()
        m.F[0].fix()
        m.T[0].fix()

        constraints = list(m.component_data_objects(pyo.Constraint))
        self.assertEqual(
            len(list(generate_strongly_connected_components(constraints))), N + 1
        )
        for i, (block, inputs) in enumerate(
            generate_strongly_connected_components(constraints)
        ):
            with TemporarySubsystemManager(to_fix=inputs):
                if i == 0:
                    # P[0], ideal_gas[0]
                    self.assertEqual(len(block.vars), 1)
                    self.assertEqual(len(block.cons), 1)

                    var_set = ComponentSet([m.P[i]])
                    con_set = ComponentSet([m.ideal_gas[i]])
                    for var, con in zip(block.vars[:], block.cons[:]):
                        self.assertIn(var, var_set)
                        self.assertIn(con, con_set)

                    # Other variables are fixed; not included
                    self.assertEqual(len(block.input_vars), 0)

                elif i == 1:
                    # P[1], rho[1], F[1], T[1], etc.
                    self.assertEqual(len(block.vars), 4)
                    self.assertEqual(len(block.cons), 4)

                    var_set = ComponentSet([m.P[i], m.rho[i], m.F[i], m.T[i]])
                    con_set = ComponentSet(
                        [m.ideal_gas[i], m.mbal[i], m.ebal[i], m.expansion[i]]
                    )
                    for var, con in zip(block.vars[:], block.cons[:]):
                        self.assertIn(var, var_set)
                        self.assertIn(con, con_set)

                    # P[0] is in expansion[1]
                    other_var_set = ComponentSet([m.P[i - 1]])
                    self.assertEqual(len(block.input_vars), 1)
                    for var in block.input_vars[:]:
                        self.assertIn(var, other_var_set)

                else:
                    # P[i], rho[i], F[i], T[i], etc.
                    self.assertEqual(len(block.vars), 4)
                    self.assertEqual(len(block.cons), 4)

                    var_set = ComponentSet([m.P[i], m.rho[i], m.F[i], m.T[i]])
                    con_set = ComponentSet(
                        [m.ideal_gas[i], m.mbal[i], m.ebal[i], m.expansion[i]]
                    )
                    for var, con in zip(block.vars[:], block.cons[:]):
                        self.assertIn(var, var_set)
                        self.assertIn(con, con_set)

                    # P[i-1], rho[i-1], F[i-1], T[i-1], etc.
                    other_var_set = ComponentSet(
                        [m.P[i - 1], m.rho[i - 1], m.F[i - 1], m.T[i - 1]]
                    )
                    self.assertEqual(len(block.input_vars), 4)
                    for var in block.input_vars[:]:
                        self.assertIn(var, other_var_set)

    def test_dynamic_backward_disc_with_initial_conditions(self):
        nfe = 5
        m = make_dynamic_model(nfe=nfe, scheme="BACKWARD")
        time = m.time
        t0 = m.time.first()
        t1 = m.time.next(t0)

        m.flow_in.fix()
        m.height[t0].fix()

        constraints = list(m.component_data_objects(pyo.Constraint))
        self.assertEqual(
            len(list(generate_strongly_connected_components(constraints))),
            nfe + 2,
            # The "initial constraints" have two SCCs because they
            # decompose into the algebraic equation and differential
            # equation. This decomposition is because the discretization
            # equation is not present.
            #
            # This is actually quite troublesome for testing because
            # it means that the topological order of strongly connected
            # components is not unique (alternatively, the initial
            # conditions and rest of the model are independent, or the
            # bipartite graph of variables and equations is disconnected).
        )
        t_scc_map = {}
        for i, (block, inputs) in enumerate(
            generate_strongly_connected_components(constraints)
        ):
            with TemporarySubsystemManager(to_fix=inputs):
                t = block.vars[0].index()
                t_scc_map[t] = i
                if t == t0:
                    continue
                else:
                    t_prev = m.time.prev(t)

                    con_set = ComponentSet(
                        [m.diff_eqn[t], m.flow_out_eqn[t], m.dhdt_disc_eq[t]]
                    )
                    var_set = ComponentSet([m.height[t], m.dhdt[t], m.flow_out[t]])
                    self.assertEqual(len(con_set), len(block.cons))
                    self.assertEqual(len(var_set), len(block.vars))
                    for var, con in zip(block.vars[:], block.cons[:]):
                        self.assertIn(var, var_set)
                        self.assertIn(con, con_set)
                        self.assertFalse(var.fixed)

                    other_var_set = (
                        ComponentSet([m.height[t_prev]]) if t != t1 else ComponentSet()
                    )
                    # At t1, "input var" height[t0] is fixed, so
                    # it is not included here.
                    self.assertEqual(len(inputs), len(other_var_set))
                    for var in block.input_vars[:]:
                        self.assertIn(var, other_var_set)
                        self.assertTrue(var.fixed)

        scc = -1
        for t in m.time:
            if t == t0:
                self.assertTrue(m.height[t].fixed)
            else:
                self.assertFalse(m.height[t].fixed)

                # Make sure "finite element blocks" in the SCC DAG are
                # in a valid topological order
                self.assertGreater(t_scc_map[t], scc)
                scc = t_scc_map[t]

            self.assertFalse(m.flow_out[t].fixed)
            self.assertFalse(m.dhdt[t].fixed)
            self.assertTrue(m.flow_in[t].fixed)

    def test_dynamic_backward_disc_without_initial_conditions(self):
        nfe = 5
        m = make_dynamic_model(nfe=nfe, scheme="BACKWARD")
        time = m.time
        t0 = m.time.first()
        t1 = m.time.next(t0)

        m.flow_in.fix()
        m.height[t0].fix()
        m.flow_out[t0].fix()
        m.dhdt[t0].fix()
        m.diff_eqn[t0].deactivate()
        m.flow_out_eqn[t0].deactivate()

        constraints = list(m.component_data_objects(pyo.Constraint, active=True))
        self.assertEqual(
            len(list(generate_strongly_connected_components(constraints))), nfe
        )
        for i, (block, inputs) in enumerate(
            generate_strongly_connected_components(constraints)
        ):
            with TemporarySubsystemManager(to_fix=inputs):
                # We have a much easier time testing the SCCs generated
                # in this test.
                t = m.time[i + 2]
                t_prev = m.time.prev(t)

                con_set = ComponentSet(
                    [m.diff_eqn[t], m.flow_out_eqn[t], m.dhdt_disc_eq[t]]
                )
                var_set = ComponentSet([m.height[t], m.dhdt[t], m.flow_out[t]])
                self.assertEqual(len(con_set), len(block.cons))
                self.assertEqual(len(var_set), len(block.vars))
                for var, con in zip(block.vars[:], block.cons[:]):
                    self.assertIn(var, var_set)
                    self.assertIn(con, con_set)
                    self.assertFalse(var.fixed)

                other_var_set = (
                    ComponentSet([m.height[t_prev]]) if t != t1 else ComponentSet()
                )
                # At t1, "input var" height[t0] is fixed, so
                # it is not included here.
                self.assertEqual(len(inputs), len(other_var_set))
                for var in block.input_vars[:]:
                    self.assertIn(var, other_var_set)
                    self.assertTrue(var.fixed)

        for t in time:
            if t == t0:
                self.assertTrue(m.height[t].fixed)
                self.assertTrue(m.flow_out[t].fixed)
                self.assertTrue(m.dhdt[t].fixed)
            else:
                self.assertFalse(m.height[t].fixed)
                self.assertFalse(m.flow_out[t].fixed)
                self.assertFalse(m.dhdt[t].fixed)

    def test_dynamic_backward_with_inputs(self):
        nfe = 5
        m = make_dynamic_model(nfe=nfe, scheme="BACKWARD")
        time = m.time
        t0 = m.time.first()
        t1 = m.time.next(t0)

        # Initial conditions are still fixed
        m.height[t0].fix()
        m.flow_out[t0].fix()
        m.dhdt[t0].fix()
        m.diff_eqn[t0].deactivate()
        m.flow_out_eqn[t0].deactivate()

        # Variables that we want in our SCCs:
        # Here we exclude "dynamic inputs" (flow_in) instead of fixing them
        variables = [
            var
            for var in m.component_data_objects(pyo.Var)
            if not var.fixed and var.parent_component() is not m.flow_in
        ]
        constraints = list(m.component_data_objects(pyo.Constraint, active=True))
        self.assertEqual(
            len(list(generate_strongly_connected_components(constraints, variables))),
            nfe,
        )

        # The result of the generator is the same as in the previous
        # test, but we are using the more general API
        for i, (block, inputs) in enumerate(
            generate_strongly_connected_components(constraints, variables)
        ):
            with TemporarySubsystemManager(to_fix=inputs):
                t = m.time[i + 2]
                t_prev = m.time.prev(t)

                con_set = ComponentSet(
                    [m.diff_eqn[t], m.flow_out_eqn[t], m.dhdt_disc_eq[t]]
                )
                var_set = ComponentSet([m.height[t], m.dhdt[t], m.flow_out[t]])
                self.assertEqual(len(con_set), len(block.cons))
                self.assertEqual(len(var_set), len(block.vars))
                for var, con in zip(block.vars[:], block.cons[:]):
                    self.assertIn(var, var_set)
                    self.assertIn(con, con_set)
                    self.assertFalse(var.fixed)

                other_var_set = ComponentSet([m.flow_in[t]])
                if t != t1:
                    other_var_set.add(m.height[t_prev])
                    # At t1, "input var" height[t0] is fixed, so
                    # it is not included here.
                self.assertEqual(len(inputs), len(other_var_set))
                for var in block.input_vars[:]:
                    self.assertIn(var, other_var_set)
                    self.assertTrue(var.fixed)

        for t in time:
            if t == t0:
                self.assertTrue(m.height[t].fixed)
                self.assertTrue(m.flow_out[t].fixed)
                self.assertTrue(m.dhdt[t].fixed)
            else:
                self.assertFalse(m.height[t].fixed)
                self.assertFalse(m.flow_out[t].fixed)
                self.assertFalse(m.dhdt[t].fixed)

    def test_dynamic_forward_disc(self):
        nfe = 5
        m = make_dynamic_model(nfe=nfe, scheme="FORWARD")
        time = m.time
        t0 = m.time.first()
        t1 = m.time.next(t0)

        m.flow_in.fix()
        m.height[t0].fix()

        constraints = list(m.component_data_objects(pyo.Constraint))
        # For a forward discretization, the entire model decomposes
        self.assertEqual(
            len(list(generate_strongly_connected_components(constraints))),
            len(list(m.component_data_objects(pyo.Constraint))),
        )
        self.assertEqual(
            len(list(generate_strongly_connected_components(constraints))),
            3 * nfe + 2,
            # "Initial constraints" only add two variables/equations
        )
        for i, (block, inputs) in enumerate(
            generate_strongly_connected_components(constraints)
        ):
            with TemporarySubsystemManager(to_fix=inputs):
                # The order is:
                #   algebraic -> derivative -> differential -> algebraic -> ...
                idx = i // 3
                mod = i % 3
                t = m.time[idx + 1]
                if t != time.last():
                    t_next = m.time.next(t)

                self.assertEqual(len(block.vars), 1)
                self.assertEqual(len(block.cons), 1)

                if mod == 0:
                    self.assertIs(block.vars[0], m.flow_out[t])
                    self.assertIs(block.cons[0], m.flow_out_eqn[t])
                elif mod == 1:
                    self.assertIs(block.vars[0], m.dhdt[t])
                    self.assertIs(block.cons[0], m.diff_eqn[t])
                elif mod == 2:
                    # Never get to mod == 2 when t == time.last()
                    self.assertIs(block.vars[0], m.height[t_next])
                    self.assertIs(block.cons[0], m.dhdt_disc_eq[t])

    def test_with_zero_coefficients(self):
        """Test where the blocks we identify are incorrect if we don't filter
        out variables with coefficients of zero
        """
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3], initialize=1.0)
        m.eq1 = pyo.Constraint(expr=m.x[1] + 2 * m.x[2] + 0 * m.x[3] == 7)
        m.eq2 = pyo.Constraint(expr=m.x[1] + pyo.log(m.x[1]) == 0)
        blocks = generate_strongly_connected_components([m.eq1, m.eq2])
        blocks = [bl for (bl, _) in blocks]

        self.assertEqual(len(blocks[0].vars), 1)
        self.assertIs(blocks[0].vars[0], m.x[1])
        self.assertIs(blocks[0].cons[0], m.eq2)

        self.assertEqual(len(blocks[1].vars), 1)
        self.assertIs(blocks[1].vars[0], m.x[2])
        self.assertIs(blocks[1].cons[0], m.eq1)


@unittest.skipUnless(scipy_available, "SciPy is not available")
@unittest.skipUnless(networkx_available, "NetworkX is not available")
class TestSolveSCC(unittest.TestCase):
    def test_dynamic_backward_no_solver(self):
        nfe = 5
        m = make_dynamic_model(nfe=nfe, scheme="BACKWARD")
        time = m.time
        t0 = time.first()
        m.flow_in.fix()
        m.height[t0].fix()

        with self.assertRaisesRegex(RuntimeError, "An external solver is required*"):
            solve_strongly_connected_components(m)

        for t in time:
            if t == t0:
                self.assertTrue(m.height[t].fixed)
            else:
                self.assertFalse(m.height[t].fixed)
            self.assertFalse(m.flow_out[t].fixed)
            self.assertFalse(m.dhdt[t].fixed)
            self.assertTrue(m.flow_in[t].fixed)

    @unittest.skipUnless(
        pyo.SolverFactory("ipopt").available(), "IPOPT is not available"
    )
    def test_dynamic_backward(self):
        nfe = 5
        m = make_dynamic_model(nfe=nfe, scheme="BACKWARD")
        time = m.time
        t0 = time.first()
        m.flow_in.fix()
        m.height[t0].fix()

        solver = pyo.SolverFactory("ipopt")
        solve_kwds = {"tee": False}
        solve_strongly_connected_components(m, solver=solver, solve_kwds=solve_kwds)

        for con in m.component_data_objects(pyo.Constraint):
            # Sanity check that this is an equality constraint...
            self.assertEqual(pyo.value(con.upper), pyo.value(con.lower))

            # Assert that the constraint is satisfied within tolerance
            self.assertAlmostEqual(
                pyo.value(con.body), pyo.value(con.upper), delta=1e-7
            )

        for t in time:
            if t == t0:
                self.assertTrue(m.height[t].fixed)
            else:
                self.assertFalse(m.height[t].fixed)
            self.assertFalse(m.flow_out[t].fixed)
            self.assertFalse(m.dhdt[t].fixed)
            self.assertTrue(m.flow_in[t].fixed)

    def test_dynamic_forward(self):
        nfe = 5
        m = make_dynamic_model(nfe=nfe, scheme="FORWARD")
        time = m.time
        t0 = time.first()
        m.flow_in.fix()
        m.height[t0].fix()

        solve_strongly_connected_components(m)

        for con in m.component_data_objects(pyo.Constraint):
            # Sanity check that this is an equality constraint...
            self.assertEqual(pyo.value(con.upper), pyo.value(con.lower))

            # Assert that the constraint is satisfied within tolerance
            self.assertAlmostEqual(
                pyo.value(con.body), pyo.value(con.upper), delta=1e-7
            )

        for t in time:
            if t == t0:
                self.assertTrue(m.height[t].fixed)
            else:
                self.assertFalse(m.height[t].fixed)
            self.assertFalse(m.flow_out[t].fixed)
            self.assertFalse(m.dhdt[t].fixed)
            self.assertTrue(m.flow_in[t].fixed)

    def test_with_calc_var_kwds(self):
        m = pyo.ConcreteModel()
        m.v0 = pyo.Var()
        m.v1 = pyo.Var()
        m.v2 = pyo.Var(initialize=79703634.05074187)
        m.v2.fix()
        m.p0 = pyo.Param(initialize=3e5)
        m.p1 = pyo.Param(initialize=1.296e12)
        m.con0 = pyo.Constraint(expr=m.v0 == m.p0)
        m.con1 = pyo.Constraint(expr=0.0 == m.p1 * m.v1 / m.v0 + m.v2)
        calc_var_kwds = {"eps": 1e-7}
        # This solve fails to converge without raising the tolerance.
        # The secant method used to converge a linear variable-constraint
        # pair (such as con0, v0 and con1, v1) leaves us with a residual of
        # 1.48e-8 due to roundoff error.
        results = solve_strongly_connected_components(m, calc_var_kwds=calc_var_kwds)
        self.assertEqual(len(results), 2)
        self.assertAlmostEqual(m.v0.value, m.p0.value)
        self.assertAlmostEqual(m.v1.value, -18.4499152895)
        # -18.4499 ~= -v2*v0/p1

    def test_with_zero_coefficients(self):
        """Test where the blocks we identify are incorrect if we don't filter
        out variables with coefficients of zero
        """
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3], initialize=1.0)
        m.eq1 = pyo.Constraint(expr=m.x[1] + 2 * m.x[2] + 0 * m.x[3] == 7)
        m.eq2 = pyo.Constraint(expr=m.x[1] + pyo.log(m.x[1]) == 0)
        results = solve_strongly_connected_components(m)

        self.assertAlmostEqual(m.x[1].value, 0.56714329)
        self.assertAlmostEqual(m.x[2].value, 3.21642835)
        self.assertEqual(m.x[3].value, 1.0)

    def test_with_inequalities(self):
        """Test that we correctly ignore inequalities"""
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3], initialize=1.0)
        m.eq1 = pyo.Constraint(expr=m.x[1] + 2 * m.x[2] + 0 * m.x[3] == 7)
        m.eq2 = pyo.Constraint(expr=m.x[1] + pyo.log(m.x[1]) == 0)

        # Solving the system violates this inequality. That is fine. We happily
        # ignore it for the purpose of this equation solve.
        m.ineq1 = pyo.Constraint(expr=m.x[1] + 2 * m.x[2] + m.x[3] <= 3)

        results = solve_strongly_connected_components(m)

        self.assertAlmostEqual(m.x[1].value, 0.56714329)
        self.assertAlmostEqual(m.x[2].value, 3.21642835)
        self.assertEqual(m.x[3].value, 1.0)


@unittest.skipUnless(scipy_available, "SciPy is not available")
@unittest.skipUnless(networkx_available, "NetworkX is not available")
class TestExceptions(unittest.TestCase):
    def test_nonsquare_system(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2], initialize=1)
        m.eq = pyo.Constraint(expr=m.x[1] + m.x[2] == 1)

        msg = "Got 2 variables and 1 constraints"
        with self.assertRaisesRegex(RuntimeError, msg):
            list(
                generate_strongly_connected_components(
                    constraints=[m.eq], variables=[m.x[1], m.x[2]]
                )
            )


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import (
    numpy,
    numpy_available as has_numpy,
    scipy,
    scipy_available as has_scipy,
)
from pyomo.common.modeling import NOTSET
from pyomo.core.expr.expr_common import _type_check_exception_arg
from pyomo.core.expr.numvalue import NumericValue, value
from pyomo.core.kernel.constraint import IConstraint, constraint_tuple

_noarg = object()
_pos_inf = float('inf')
_neg_inf = float('-inf')

#
# Note: This class is experimental. The implementation may
#       change or it may go away.
#


class _MatrixConstraintData(IConstraint):
    """
    A placeholder object for linear constraints in a
    matrix_constraint container. A user should not
    directly instantiate this class.
    """

    _ctype = IConstraint
    _linear_canonical_form = True
    __slots__ = ("_parent", "_storage_key", "_active", "__weakref__")

    def __init__(self, index):
        assert index >= 0
        self._parent = None
        self._storage_key = index
        self._active = True

    @property
    def index(self):
        """The row index of this constraint in the parent matrix"""
        return self._storage_key

    @property
    def terms(self):
        """An iterator over the terms in the body of this
        constraint as (variable, coefficient) tuples"""
        parent = self.parent
        x = parent.x
        if x is None:
            raise ValueError("No variable order has been assigned")
        A = parent._A
        if parent._sparse:
            for k in range(
                A.indptr[self._storage_key], A.indptr[self._storage_key + 1]
            ):
                yield x[A.indices[k]], A.data[k]
        else:
            for item in zip(x, A[self._storage_key, :].tolist()):
                yield item

    #
    # Override a the default __call__ method on IConstraint
    # to avoid building the body expression
    #

    def __call__(self, exception=NOTSET):
        # don't mask an exception in the terms
        # property method
        exception = _type_check_exception_arg(self, exception)
        if self.parent.x is None:
            raise ValueError("No variable order has been assigned")
        try:
            return sum(c * v() for v, c in self.terms)
        except (ValueError, TypeError):
            if exception:
                raise
            return None

    #
    # Define the IConstraint abstract methods
    #

    @property
    def body(self):
        """The body of the constraint"""
        return sum(c * v for v, c in self.terms)

    @property
    def lower(self):
        """The expression for the lower bound of the constraint"""
        return self.parent.lb[self._storage_key]

    @lower.setter
    def lower(self, lb):
        if self.equality:
            raise ValueError(
                "The lower property can not be set "
                "when the equality property is True."
            )
        if lb is None:
            lb = -numpy.inf
        elif isinstance(lb, NumericValue):
            raise ValueError("lb must be set to a simple numeric type or None")
        self.parent.lb[self._storage_key] = lb

    @property
    def upper(self):
        """The expression for the upper bound of the constraint"""
        return self.parent.ub[self._storage_key]

    @upper.setter
    def upper(self, ub):
        if self.equality:
            raise ValueError(
                "The upper property can not be set "
                "when the equality property is True."
            )
        if ub is None:
            ub = numpy.inf
        elif isinstance(ub, NumericValue):
            raise ValueError("ub must be set to a simple numeric type or None")
        self.parent.ub[self._storage_key] = ub

    @property
    def lb(self):
        """The value of the lower bound of the constraint"""
        lb = value(self.lower)
        if lb == _neg_inf:
            return None
        return lb

    @lb.setter
    def lb(self, lb):
        self.lower = lb

    @property
    def ub(self):
        """The value of the upper bound of the constraint"""
        ub = value(self.upper)
        if ub == _pos_inf:
            return None
        return ub

    @ub.setter
    def ub(self, ub):
        self.upper = ub

    @property
    def rhs(self):
        """The right-hand side of the constraint. This
        property can only be read when the equality property
        is :const:`True`. Assigning to this property
        implicitly sets the equality property to
        :const:`True`."""
        if not self.equality:
            raise ValueError(
                "The rhs property can not be read "
                "when the equality property is False."
            )
        return self.parent.lb[self._storage_key]

    @rhs.setter
    def rhs(self, rhs):
        if rhs is None:
            # None has a different meaning depending on the
            # context (lb or ub), so there is no way to
            # interpret this
            raise ValueError(
                "Constraint right-hand side can not be assigned a value of None."
            )
        elif isinstance(rhs, NumericValue):
            raise ValueError("rhs must be set to a simple numeric type or None")
        self.parent.lb[self._storage_key] = rhs
        self.parent.ub[self._storage_key] = rhs
        self.parent.equality[self._storage_key] = True

    @property
    def bounds(self):
        """The bounds of the constraint as a tuple (lb, ub)"""
        return (self.parent.lb[self._storage_key], self.parent.ub[self._storage_key])

    @bounds.setter
    def bounds(self, bounds_tuple):
        self.lb, self.ub = bounds_tuple

    @property
    def equality(self):
        """Returns :const:`True` when this is an equality
        constraint.

        Disable equality by assigning
        :const:`False`. Equality can only be activated by
        assigning a value to the .rhs property."""
        return self.parent.equality[self._storage_key]

    @equality.setter
    def equality(self, equality):
        if equality:
            raise ValueError(
                "The constraint equality flag can "
                "only be set to True by assigning "
                "a value to the rhs property "
                "(e.g., con.rhs = con.lb)."
            )
        assert not equality
        self.parent.equality[self._storage_key] = False

    #
    # Define methods that writers expect when the
    # _linear_canonical_form flag is True
    #

    def canonical_form(self, compute_values=True):
        """Build a canonical representation of the body of
        this constraints"""
        from pyomo.repn.standard_repn import StandardRepn

        variables = []
        coefficients = []
        constant = 0
        for v, c in self.terms:
            # we call float to get rid of the numpy type
            c = float(c)
            if not v.fixed:
                variables.append(v)
                coefficients.append(c)
            else:
                if compute_values:
                    constant += c * v()
                else:
                    constant += c * v
        repn = StandardRepn()
        repn.linear_vars = tuple(variables)
        repn.linear_coefs = tuple(coefficients)
        repn.constant = constant
        return repn


class matrix_constraint(constraint_tuple):
    """
    A container for constraints of the form lb <= Ax <= ub.

    Args:
        A: A scipy sparse matrix or 2D numpy array (always
            copied)
        lb: A scalar or array with the same number of rows
            as A that defines the lower bound of the
            constraints
        ub: A scalar or array with the same number of rows
            as A that defines the upper bound of the
            constraints
        rhs: A scalar or array with the same number of rows
            as A that defines the right-hand side of the
            constraints (implies equality constraints)
        x: A list with the same number of columns as A that
            stores the variable associated with each column
        sparse: Indicates whether or not sparse storage (CSR
            format) should be used to store A. Default is
            :const:`True`.
    """

    __slots__ = ("_A", "_sparse", "_lb", "_ub", "_equality", "_x")

    def __init__(self, A, lb=None, ub=None, rhs=None, x=None, sparse=True):
        if (not has_numpy) or (not has_scipy):  # pragma:nocover
            raise ValueError("This class requires numpy and scipy")

        m, n = A.shape
        assert m > 0
        assert n > 0
        cons = (_MatrixConstraintData(i) for i in range(m))
        super(matrix_constraint, self).__init__(cons)

        if sparse:
            self._sparse = True
            self._A = scipy.sparse.csr_matrix(A, dtype=float, copy=True)
            self._A.data.setflags(write=False)
            self._A.indices.setflags(write=False)
            self._A.indptr.setflags(write=False)
        else:
            self._sparse = False
            self._A = numpy.array(A, dtype=float, copy=True)
            self._A.setflags(write=False)
        self._lb = numpy.ndarray(m, dtype=float)
        self._ub = numpy.ndarray(m, dtype=float)
        self._equality = numpy.ndarray(m, dtype=bool)
        self._equality.fill(False)

        # now use the setters to fill the arrays
        self.x = x
        if rhs is None:
            self.lb = lb
            self.ub = ub
        else:
            if (lb is not None) or (ub is not None):
                raise ValueError(
                    "The 'rhs' keyword can not "
                    "be used with the 'lb' or "
                    "'ub' keywords to initialize"
                    " a constraint."
                )
            self.rhs = rhs

    @property
    def sparse(self):
        """Boolean indicating whether or not the underlying
        matrix uses sparse storage"""
        return self._sparse

    @property
    def A(self):
        """A read-only view of the constraint matrix"""
        if self._sparse:
            return scipy.sparse.csr_matrix(self._A, copy=False)
        else:
            return self._A.view()

    @property
    def x(self):
        """The list of variables associated with the columns
        of the constraint matrix"""
        return self._x

    @x.setter
    def x(self, x):
        if x is None:
            self._x = None
        else:
            x = tuple(x)
            m, n = self._A.shape
            if len(x) != n:
                raise ValueError("Argument length must be %s not %s" % (n, len(x)))
            self._x = x

    @property
    def lb(self):
        """The array of constraint lower bounds"""
        return self._lb.view()

    @lb.setter
    def lb(self, lb):
        if self.equality.any():
            raise ValueError(
                "The lb array can not be set "
                "when there are indices of the "
                "equality array that are True"
            )
        if lb is None:
            lb = -numpy.inf
        if isinstance(lb, numpy.ndarray):
            numpy.copyto(self._lb, lb)
        elif isinstance(lb, NumericValue):
            raise ValueError("lb must be set to a simple numeric type or a numpy array")
        else:
            self._lb.fill(lb)

    @property
    def ub(self):
        """The array of constraint upper bounds"""
        return self._ub.view()

    @ub.setter
    def ub(self, ub):
        if self.equality.any():
            raise ValueError(
                "The ub array can not be set "
                "when there are indices of the "
                "equality array that are True"
            )
        if ub is None:
            ub = numpy.inf
        if isinstance(ub, numpy.ndarray):
            numpy.copyto(self._ub, ub)
        elif isinstance(ub, NumericValue):
            raise ValueError("ub must be set to a simple numeric type or a numpy array")
        else:
            self._ub.fill(ub)

    @property
    def rhs(self):
        """The array of constraint right-hand sides. Can be
        set to a scalar or a numpy array of the same
        dimension. This property can only be read when the
        equality property is :const:`True` on every
        index. Assigning to this property implicitly sets
        the equality property to :const:`True` on every
        index."""
        if not self.equality.all():
            raise ValueError(
                "The rhs array can not be read when "
                "there are indices of the equality array "
                "that are False."
            )
        return self._lb.view()

    @rhs.setter
    def rhs(self, rhs):
        if rhs is None:
            # None has a different meaning depending on the
            # context (lb or ub), so there is no way to
            # interpret this
            raise ValueError(
                "Constraint right-hand side can not be assigned a value of None."
            )
        elif isinstance(rhs, NumericValue):
            raise ValueError(
                "rhs must be set to a simple numeric type or a numpy array"
            )
        elif isinstance(rhs, numpy.ndarray):
            numpy.copyto(self._lb, rhs)
            numpy.copyto(self._ub, rhs)
        else:
            self._lb.fill(rhs)
            self._ub.fill(rhs)
        self._equality.fill(True)

    @property
    def equality(self):
        """The array of boolean entries indicating the
        indices that are equality constraints"""
        return self._equality.view()

    @equality.setter
    def equality(self, equality):
        if equality:
            raise ValueError(
                "The constraint equality flag can "
                "only be set to True by assigning "
                "an expression to the rhs property "
                "(e.g., con.rhs = con.lb)."
            )
        assert not equality
        self._equality.fill(False)

    def __call__(self, exception=NOTSET):
        """Compute the value of the body of this constraint"""
        exception = _type_check_exception_arg(self, exception)
        if self.x is None:
            raise ValueError("No variable order has been assigned")
        values = numpy.array([v.value for v in self.x], dtype=float)
        if numpy.isnan(values).any():
            if exception:
                raise ValueError("One or more variables do not have a value")
            return None
        return self._A.dot(values)

    @property
    def lslack(self, body=_noarg):
        """Lower slack (body - lb)"""
        # this method is written so that constraint
        # types that build the body expression on the
        # fly do not have to here
        if body is _noarg:
            body = self(exception=False)
        if body is None:
            return None
        return body - self.lb

    @property
    def uslack(self, body=_noarg):
        """Upper slack (ub - body)"""
        # this method is written so that constraint
        # types that build the body expression on the
        # fly do not have to here
        if body is _noarg:
            body = self(exception=False)
        if body is None:
            return None
        return self.ub - body

    @property
    def slack(self):
        """min(lslack, uslack)"""
        # this method is written so that constraint
        # types that build the body expression on the
        # fly do not have to here
        body = self(exception=False)
        if body is None:
            return None
        lslack = self.__class__.lslack.fget(self, body=body)
        uslack = self.__class__.uslack.fget(self, body=body)
        return numpy.minimum(lslack, uslack)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os
import shutil
import tempfile

import pyomo.common.envvar as envvar

from pyomo.common.fileutils import this_file_dir, find_dir
from pyomo.common.download import FileDownloader


def _generate_configuration():
    # defer the import until use (this eventually imports pkg_resources,
    # which is slow to import)
    from setuptools.extension import Extension

    # Try and find MC++.  Defer to the MCPP_ROOT if it is set;
    # otherwise, look in common locations for a mcpp directory.
    pathlist = [os.path.join(envvar.PYOMO_CONFIG_DIR, 'src'), this_file_dir()]
    if 'MCPP_ROOT' in os.environ:
        mcpp = os.environ['MCPP_ROOT']
    else:
        mcpp = find_dir('mcpp', cwd=True, pathlist=pathlist)
    if mcpp:
        print("Found MC++ at %s" % (mcpp,))
    else:
        raise RuntimeError(
            "Cannot identify the location of the MCPP source distribution"
        )

    #
    # Configuration for this extension
    #
    project_dir = this_file_dir()
    sources = [os.path.join(project_dir, 'mcppInterface.cpp')]
    include_dirs = [
        os.path.join(mcpp, 'src', 'mc'),
        os.path.join(mcpp, 'src', '3rdparty', 'fadbad++'),
    ]

    mcpp_ext = Extension(
        "mcppInterface",
        sources=sources,
        language="c++",
        extra_compile_args=[],
        include_dirs=include_dirs,
        library_dirs=[],
        libraries=[],
    )

    package_config = {'name': 'mcpp', 'packages': [], 'ext_modules': [mcpp_ext]}

    return package_config


def build_mcpp():
    from setuptools import Distribution
    from setuptools.command.build_ext import build_ext

    class _BuildWithoutPlatformInfo(build_ext, object):
        # Python3.x puts platform information into the generated SO file
        # name, which is usually fine for python extensions, but since this
        # is not a "real" extension, we will hijack things to remove the
        # platform information from the filename so that Pyomo can more
        # easily locate it.  Note that build_ext is not a new-style class in
        # Python 2.7, so we will add an explicit inheritance from object so
        # that super() works.
        def get_ext_filename(self, ext_name):
            filename = (
                super(_BuildWithoutPlatformInfo, self)
                .get_ext_filename(ext_name)
                .split('.')
            )
            filename = '.'.join([filename[0], filename[-1]])
            return filename

    print("\n**** Building MCPP library ****")
    package_config = _generate_configuration()
    package_config['cmdclass'] = {'build_ext': _BuildWithoutPlatformInfo}
    dist = Distribution(package_config)
    install_dir = os.path.join(envvar.PYOMO_CONFIG_DIR, 'lib')
    dist.get_command_obj('install_lib').install_dir = install_dir
    try:
        basedir = os.path.abspath(os.path.curdir)
        tmpdir = os.path.abspath(tempfile.mkdtemp())
        print("   tmpdir = %s" % (tmpdir,))
        os.chdir(tmpdir)
        dist.run_command('install_lib')
        print("Installed mcppInterface to %s" % (install_dir,))
    finally:
        os.chdir(basedir)
        shutil.rmtree(tmpdir)


class MCPPBuilder:
    def __call__(self, parallel):
        return build_mcpp()

    def skip(self):
        return FileDownloader.get_sysinfo()[0] == 'windows'


if __name__ == "__main__":
    build_mcpp()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Transformation from BooleanVar and LogicalConstraint to Binary and
Constraints."""

from pyomo.common.collections import ComponentMap
from pyomo.common.errors import MouseTrap, DeveloperError
from pyomo.common.modeling import unique_component_name
from pyomo.common.config import ConfigBlock, ConfigValue
from pyomo.contrib.fbbt.fbbt import compute_bounds_on_expr
from pyomo.core import (
    TransformationFactory,
    BooleanVar,
    VarList,
    Binary,
    LogicalConstraint,
    Block,
    ConstraintList,
    native_types,
    BooleanVarList,
    SortComponents,
)
from pyomo.core.base.block import BlockData
from pyomo.core.base.boolean_var import _DeprecatedImplicitAssociatedBinaryVariable
from pyomo.core.expr.cnf_walker import to_cnf
from pyomo.core.expr import (
    AndExpression,
    OrExpression,
    NotExpression,
    AtLeastExpression,
    AtMostExpression,
    ExactlyExpression,
    special_boolean_atom_types,
    EqualityExpression,
    InequalityExpression,
    RangedExpression,
    identify_variables,
)
from pyomo.core.expr.numvalue import native_logical_types, value
from pyomo.core.expr.visitor import StreamBasedExpressionVisitor
from pyomo.core.plugins.transform.hierarchy import IsomorphicTransformation
from pyomo.core.util import target_list


@TransformationFactory.register(
    "core.logical_to_linear", doc="Convert logic to linear constraints"
)
class LogicalToLinear(IsomorphicTransformation):
    """
    Re-encode logical constraints as linear constraints,
    converting Boolean variables to binary.
    """

    CONFIG = ConfigBlock('core.logical_to_linear')
    CONFIG.declare(
        'targets',
        ConfigValue(
            default=None,
            domain=target_list,
            description="target or list of targets that will be relaxed",
            doc="""
            This specifies the list of LogicalConstraints to transform, or the
            list of Blocks or Disjuncts on which to transform all of the
            LogicalConstraints. Note that if the transformation is done out
            of place, the list of targets should be attached to the model before it
            is cloned, and the list will specify the targets on the cloned
            instance.
            """,
        ),
    )

    def _apply_to(self, model, **kwds):
        config = self.CONFIG(kwds.pop('options', {}))
        config.set_value(kwds)
        targets = config.targets
        if targets is None:
            targets = (model,)

        new_var_lists = ComponentMap()
        transBlocks = {}
        for t in targets:
            # If the user promises that the target is Block-like, we will go
            # with it. Note, however, that they can only use targets for
            # this--when we go searching for stuff to transform we will only
            # look on Blocks. And yes, this means we are ignoring Disjuncts. We
            # are in fact ignoring all GDP components because this
            # transformation is a promise only to transform LogicalConstraints
            # and the relevant BooleanVars, not to create an algebraic
            # model. (We are making this decision largely because having this
            # transformation do anything to GDP stuff is an assumption on how
            # the GDP will be solved, and it would be wrong to assume that a GDP
            # will *necessarily* be solved as an algebraic model. The star
            # example of not doing so being GDPopt.)
            if t.ctype is Block or isinstance(t, BlockData):
                self._transform_block(t, model, new_var_lists, transBlocks)
            elif t.ctype is LogicalConstraint:
                if t.is_indexed():
                    self._transform_constraint(t, new_var_lists, transBlocks)
                else:
                    self._transform_constraintData(t, new_var_lists, transBlocks)
            else:
                raise RuntimeError(
                    "Target '%s' was not a Block, Disjunct, or"
                    " LogicalConstraint. It was of type %s "
                    "and can't be transformed." % (t.name, type(t))
                )

    def _transform_boolean_varData(self, bool_vardata, new_varlists):
        # This transformation tries to group the binaries it creates for indexed
        # BooleanVars onto the same VarList. This won't work across separate
        # calls to the transformation, but within one call it's fine. So we have
        # two cases: 1) either we have created a VarList for this
        # BooleanVarData's parent_component, but have yet to add its binary to
        # said list, or 2) we have neither the binary nor the VarList

        parent_component = bool_vardata.parent_component()
        new_varlist = new_varlists.get(parent_component)
        if new_varlist is None and bool_vardata.get_associated_binary() is None:
            # Case 2) we have neither the VarList nor an associated binary
            parent_block = bool_vardata.parent_block()
            new_var_list_name = unique_component_name(
                parent_block, parent_component.local_name + '_asbinary'
            )
            new_varlist = VarList(domain=Binary)
            setattr(parent_block, new_var_list_name, new_varlist)
            new_varlists[parent_component] = new_varlist

        if bool_vardata.get_associated_binary() is None:
            # Case 1) we already have a VarList, but need to create the
            # associated binary
            new_binary_vardata = new_varlist.add()
            bool_vardata.associate_binary_var(new_binary_vardata)
            if bool_vardata.value is not None:
                new_binary_vardata.value = int(bool_vardata.value)
            if bool_vardata.fixed:
                new_binary_vardata.fix()

    def _transform_constraint(self, constraint, new_varlists, transBlocks):
        for i in constraint.keys(sort=SortComponents.ORDERED_INDICES):
            self._transform_constraintData(constraint[i], new_varlists, transBlocks)
        constraint.deactivate()

    def _transform_block(self, target_block, model, new_varlists, transBlocks):
        _blocks = (
            target_block.values() if target_block.is_indexed() else (target_block,)
        )
        for block in _blocks:
            for logical_constraint in block.component_objects(
                ctype=LogicalConstraint, active=True, descend_into=Block
            ):
                self._transform_constraint(
                    logical_constraint, new_varlists, transBlocks
                )

            # This can go away when we deprecate this transformation
            # transforming BooleanVars. This just marks the BooleanVars as
            # "seen" so that if someone asks for their binary var later, we can
            # create it on the fly and complain.
            for bool_vardata in block.component_data_objects(
                BooleanVar, descend_into=Block
            ):
                if bool_vardata._associated_binary is None:
                    bool_vardata._associated_binary = (
                        _DeprecatedImplicitAssociatedBinaryVariable(bool_vardata)
                    )

    def _transform_constraintData(self, logical_constraint, new_varlists, transBlocks):
        # first find all the relevant BooleanVars and associate a binary (if
        # they don't have one already)
        for bool_vardata in identify_variables(logical_constraint.expr):
            if bool_vardata.ctype is BooleanVar:
                self._transform_boolean_varData(bool_vardata, new_varlists)

        # now create a transformation block on the constraint's parent block (if
        # we don't have one already)
        parent_block = logical_constraint.parent_block()
        xfrm_block = transBlocks.get(parent_block)
        if xfrm_block is None:
            xfrm_block = self._create_transformation_block(parent_block)
            transBlocks[parent_block] = xfrm_block
        new_constrlist = xfrm_block.transformed_constraints
        new_boolvarlist = xfrm_block.augmented_vars
        new_varlist = xfrm_block.augmented_vars_asbinary

        old_boolvarlist_length = len(new_boolvarlist)

        indicator_map = ComponentMap()
        cnf_statements = to_cnf(logical_constraint.body, new_boolvarlist, indicator_map)
        logical_constraint.deactivate()

        # Associate new Boolean vars to new binary variables
        num_new = len(new_boolvarlist) - old_boolvarlist_length
        list_o_vars = list(new_boolvarlist.values())
        if num_new:
            for bool_vardata in list_o_vars[-num_new:]:
                new_binary_vardata = new_varlist.add()
                bool_vardata.associate_binary_var(new_binary_vardata)

        # Add constraints associated with each CNF statement
        for cnf_statement in cnf_statements:
            for linear_constraint in _cnf_to_linear_constraint_list(cnf_statement):
                new_constrlist.add(expr=linear_constraint)

        # Add bigM associated with special atoms
        # Note: this ad-hoc reformulation may be revisited for tightness in the
        # future.
        old_varlist_length = len(new_varlist)
        for indicator_var, special_atom in indicator_map.items():
            for linear_constraint in _cnf_to_linear_constraint_list(
                special_atom, indicator_var, new_varlist
            ):
                new_constrlist.add(expr=linear_constraint)

        # Previous step may have added auxiliary binaries. Associate augmented
        # Booleans to them.
        num_new = len(new_varlist) - old_varlist_length
        list_o_vars = list(new_varlist.values())
        if num_new:
            for binary_vardata in list_o_vars[-num_new:]:
                new_bool_vardata = new_boolvarlist.add()
                new_bool_vardata.associate_binary_var(binary_vardata)

    def _create_transformation_block(self, context):
        new_xfrm_block_name = unique_component_name(context, 'logic_to_linear')
        new_xfrm_block = Block(doc="Transformation objects for logic_to_linear")
        setattr(context, new_xfrm_block_name, new_xfrm_block)

        new_xfrm_block.transformed_constraints = ConstraintList()
        new_xfrm_block.augmented_vars = BooleanVarList()
        new_xfrm_block.augmented_vars_asbinary = VarList(domain=Binary)

        return new_xfrm_block


def update_boolean_vars_from_binary(model, integer_tolerance=1e-5):
    """Updates all Boolean variables based on the value of their linked binary
    variables."""
    for boolean_var in model.component_data_objects(BooleanVar, descend_into=Block):
        binary_var = boolean_var.get_associated_binary()
        if binary_var is not None and binary_var.value is not None:
            if abs(binary_var.value - 1) <= integer_tolerance:
                boolean_var.value = True
            elif abs(binary_var.value) <= integer_tolerance:
                boolean_var.value = False
            else:
                raise ValueError(
                    "Binary variable has non-{0,1} value: "
                    "%s = %s" % (binary_var.name, binary_var.value)
                )
            boolean_var.stale = binary_var.stale


def _cnf_to_linear_constraint_list(cnf_expr, indicator_var=None, binary_varlist=None):
    # Screen for constants
    if type(cnf_expr) in native_types or cnf_expr.is_constant():
        if value(cnf_expr) is True:
            # Trivially feasible: no constraints
            return []
        else:
            # Trivially infeasible: we will return an infeasible
            # constant expression, because if we are nested within
            # something like a Disjunct, the model may still be feasible
            # (only this disjunct is not feasible).
            return [InequalityExpression((1, 0), False)]
    if cnf_expr.is_expression_type():
        return CnfToLinearVisitor(indicator_var, binary_varlist).walk_expression(
            cnf_expr
        )
    else:
        return [cnf_expr.get_associated_binary() == 1]  # Assume that cnf_expr
        # is a BooleanVar


_numeric_relational_types = {InequalityExpression, EqualityExpression, RangedExpression}


class CnfToLinearVisitor(StreamBasedExpressionVisitor):
    """Convert CNF logical constraint to linear constraints.

    Expected expression node types: AndExpression, OrExpression, NotExpression,
    AtLeastExpression, AtMostExpression, ExactlyExpression, BooleanVarData

    """

    def __init__(self, indicator_var, binary_varlist):
        super(CnfToLinearVisitor, self).__init__()
        self._indicator = indicator_var
        self._binary_varlist = binary_varlist

    def exitNode(self, node, values):
        if type(node) == AndExpression:
            return list(
                (v if type(v) in _numeric_relational_types else v == 1) for v in values
            )
        elif type(node) == OrExpression:
            return sum(values) >= 1
        elif type(node) == NotExpression:
            return 1 - values[0]
        # Note: the following special atoms should only be encountered as root
        # nodes.  If they are encountered otherwise, something went wrong.
        sum_values = sum(values[1:])
        num_args = node.nargs() - 1  # number of logical arguments
        if self._indicator is None:
            if type(node) == AtLeastExpression:
                return sum_values >= values[0]
            elif type(node) == AtMostExpression:
                return sum_values <= values[0]
            elif type(node) == ExactlyExpression:
                return sum_values == values[0]
        else:
            rhs_lb, rhs_ub = compute_bounds_on_expr(values[0])
            if rhs_lb == float('-inf') or rhs_ub == float('inf'):
                raise ValueError(
                    "Cannot generate linear constraints for %s"
                    "([N, *logical_args]) with unbounded N. "
                    "Detected %s <= N <= %s." % (type(node).__name__, rhs_lb, rhs_ub)
                )
            indicator_binary = self._indicator.get_associated_binary()
            if type(node) == AtLeastExpression:
                return [
                    sum_values >= values[0] - rhs_ub * (1 - indicator_binary),
                    sum_values
                    <= values[0] - 1 + (-(rhs_lb - 1) + num_args) * indicator_binary,
                ]
            elif type(node) == AtMostExpression:
                return [
                    sum_values
                    <= values[0] + (-rhs_lb + num_args) * (1 - indicator_binary),
                    sum_values >= (values[0] + 1) - (rhs_ub + 1) * indicator_binary,
                ]
            elif type(node) == ExactlyExpression:
                less_than_binary = self._binary_varlist.add()
                more_than_binary = self._binary_varlist.add()
                return [
                    sum_values
                    <= values[0] + (-rhs_lb + num_args) * (1 - indicator_binary),
                    sum_values >= values[0] - rhs_ub * (1 - indicator_binary),
                    indicator_binary + less_than_binary + more_than_binary >= 1,
                    sum_values
                    <= values[0]
                    - 1
                    + (-(rhs_lb - 1) + num_args) * (1 - less_than_binary),
                    sum_values >= values[0] + 1 - (rhs_ub + 1) * (1 - more_than_binary),
                ]
        if type(node) in _numeric_relational_types:
            raise MouseTrap(
                "core.logical_to_linear does not support transforming "
                "LogicalConstraints with embedded relational expressions.  "
                f"Found '{node}'."
            )
        else:
            raise DeveloperError(
                f"Unsupported node type {type(node)} encountered when "
                f"transforming a CNF expression to its linear equivalent ({node})."
            )

    def beforeChild(self, node, child, child_idx):
        if type(node) in special_boolean_atom_types and child is node.args[0]:
            return False, child
        if type(child) in native_logical_types:
            return False, int(child)
        if type(child) in native_types:
            return False, child

        if child.is_expression_type():
            return True, None

        # Only thing left should be BooleanVarData
        #
        # TODO: After the expr_multiple_dispatch is merged, this should
        # be switched to using as_numeric.
        if hasattr(child, 'get_associated_binary'):
            return False, child.get_associated_binary()
        else:
            return False, child

    def finalizeResult(self, result):
        if type(result) is list:
            return result
        elif type(result) in _numeric_relational_types:
            return [result]
        else:
            return [result == 1]
